function [] = Sec_003()
%--------------------------------------------------------------------------
% Fit a polynomial to data generated from an nth order polynomial with
% fixed measurement noise using a code, P-DATAN, structured like DATAN
%--------------------------------------------------------------------------
% 
% My general purpose Bayesian Inference code has three levels:
%
%  Top:     fun_MAP.m       iterates to the MAP point and calculates all the likelihoods 
%  Middle:  fun_JdJd2J.m    calculates the negative log likelihood and its derivatives w.r.t. the parameters and noise   
%  Bottom:  fun_solve_*.m   calculates the model output and its derivatives w.r.t the parameters 
%  
% In DATAN, the bottom level is a thermoacoustic network model. 
%
% In P-DATAN, the bottom level is the polynomial evaluation used in
% Sec_001.m
%
% The purpose of this code, Sec_003.m, is to show that its results exactly
% match those from Sec_001.m, which was created with the simple polynomial
% code. P-DATAN is, however, more versatile than the simply polynomial
% code. For example, it can optimize the measurement noise as well as the
% parameters. 

%--------------------------------------------------------------------------
% Set the model (a.k.a. the order of the polynomial)
%--------------------------------------------------------------------------
N_mod = 5;

%--------------------------------------------------------------------------
% Set the plot options
%--------------------------------------------------------------------------
plot_options.plot_gt        = true; % plot ground truth polynomial
plot_options.plot_data      = true; % plot data 
plot_options.plot_pred      = true; % plot model predictions
plot_options.print_to_file  = true; % save to file

% Open a figure and initialize plot_options if they are required
if plot_options.plot_gt || plot_options.plot_data || plot_options.plot_pred
    figure(2); clf; hold on
    plot_options.plot_handles = [];
    plot_options.plot_labels  = {};
end
% Initialize x_plt if it is required
if plot_options.plot_gt || plot_options.plot_pred
    x_plt = -0.5:0.01:0.5; n_plt = numel(x_plt);
end

%--------------------------------------------------------------------------
% Generate ground truth parameters, a_gt, from a polynomial of a user-defined order
%--------------------------------------------------------------------------
% Set the order of the ground truth polynomial
N_gt    = 5;
% Reset the random seed (optional)
rng(1);
% Generate some random data 
x_ref = -0.5:0.05:0.5; n_ref = numel(x_ref); y_ref = randn(n_ref,1);
% Find the coefficients of an order N_gt polynomial that fits this data
a_gt = polyfit(x_ref,y_ref,N_gt);
% Plot ground truth
if plot_options.plot_gt
    new_handle = plot(x_plt,polyval(a_gt,x_plt),'r-');
    new_label  = 'ground truth';
    plot_options = fun_update_legend(plot_options,new_handle,new_label);
end

%--------------------------------------------------------------------------
% Generate N_exp synthetic datapoints from the ground truth polynomial
%--------------------------------------------------------------------------
%
% Set the number of experimental datapoints
N_e  = 10;
% Set the x-locations at which the datapoints are selected
x_e  = linspace(-0.5,0.5,N_e);
% Set the variance of the measurement noise
C_ee = 0.0005;
% Sample the polynomial at the measurement points, corrupted by Gaussian noise 
y_e  = polyval(a_gt,x_e) + sqrt(C_ee)*randn(1,N_e);
% Plot data
if plot_options.plot_data
    new_handle = errorbar(x_e,y_e,2*sqrt(C_ee)*ones(N_e,1),'k.');
    new_label  = 'experimental data';
    plot_options = fun_update_legend(plot_options,new_handle,new_label);
end

%--------------------------------------------------------------------------
% Convert datapoints to P-DATAN format
%--------------------------------------------------------------------------
% Each experimental configuration is held in a structure: param{1} to
% param{M}. This structure contains the fixed parameters and the variable
% parameters. Here, we set the fixed parameters for each experimental
% configuration, which is simply the x-value of the datapoint. (There are
% no other parameters in this case.)
%
% For each experimental configuration there is some experimental data: z{1}
% to z{M}. There can be several datapoints for each configuration so z{m}
% can be a scalar (1 point) or a vector (n points). Here, we set the
% experimental datapoint for each configuration, which is simply the
% y-value of the datapoint. 

% Pre-allocate
param = cell(1,N_e); z = cell(1,N_e);
% Put into P-DATAN format
for n_e = 1:N_e
    param{n_e}.x = x_e(n_e);
    z{n_e} = y_e(n_e);
end

%--------------------------------------------------------------------------
% Set the prior and the prior covariances
%--------------------------------------------------------------------------
% 
% The prior and the prior covariances have a strong influence on the
% marginal likelihood. For example, see MacKay (1992) on how to compare
% models with different priors. 
%
% Here, we will set the priors expected values and variances to those found
% from Monte Carlo sampling of the above ground truth parameters, performed
% in prior_MC.m. This is a brute force solution, but is quick. It gives:
%
%  mean = zero
%  var  ~ exp(2*(N_mod:-1:0)), where N_mod is the model polynomial's order
% 
% Other covariances can be set with the switch below. 
%
% MacKay (1992) 'Bayesian Interpolation'; Neural Computation 4, 415--447

% Set the type of prior
prior = 'zero';
% Set the type of covariance
prior_covariance = 'std_propto_exp_n';

% Set the prior
switch prior
    case 'zero'
        a_prior_unnormalized = zeros(1,N_mod+1);
    otherwise
        disp('prior not set'); return
end

% Set the covariance
switch prior_covariance
    case 'std_propto_exp_n'
        % Approx. match to MC sample of priors
        C_aa_prior_unnormalized = diag(exp(2*(N_mod:-1:0)));
    case 'propto_n'
        % covariance scales with 2^(n), where n is the parameter multiplied by x^n. 
        C_aa_prior_unnormalized = 2 * diag(2.^(N_mod:-1:0));
    case 'loose_propto_n'
        C_aa_prior_unnormalized = 2^4 * diag(2.^(N_mod:-1:0));
    case 'very_loose_propto_n'
        C_aa_prior_unnormalized = 2^8 * diag(2.^(N_mod:-1:0));
    case 'unity'
        % covariance is the same for all coefficients
        C_aa_prior_unnormalized = eye(N_mod+1);
    otherwise
        disp('prior covariance not set'); return
end

%--------------------------------------------------------------------------
% Set the reference values for the parameters and noise
%--------------------------------------------------------------------------
% DATAN allows the user to specify reference values for the parameters so
% that, during the optimization process, all parameters have order 1. 
%
% e.g.  a_actual = a_ref * a_optimized
%
% The same is true for the measurement noise, which can also be optimized in DATAN 
%
% e.g.  C_ee_actual = C_ee_ref * C_ee_optimized
%
% Finally, a needs to be a column vector, not a row vector
switch prior_covariance
    case 'std_propto_exp_n'
        a_ref = exp((N_mod:-1:0)).'; % col vector
end
% Normalize a_prior and transpose to col vector
a_prior = a_prior_unnormalized.' ./ a_ref;
% Normalize C_aa_prior (which should become eye(N+1))
C_aa_prior = C_aa_prior_unnormalized .* ( (1./a_ref) * (1./a_ref)' );
% Normalize C_ee
[C_ee,C_ee_ref] = gen_fun_normalize(C_ee);

%--------------------------------------------------------------------------
% Calculate the MAP points, the Marginal Likelihood, the BFL and the OF
%--------------------------------------------------------------------------
%
% Set options for fun_MAP
MAP_options.reconverge_to_a_MAP_varying_a                       = true;  % <-- allow parameters to float  
MAP_options.reconverge_to_a_MAP_varying_inv_C_ee                = false; % <-- do not allow measurement noise to float
MAP_options.check_is_MAP_point                                  = false;
MAP_options.force_exact_inv_C_aa_MAP                            = false;
MAP_options.return_inv_C_aa_MAP                                 = false;
MAP_options.return_log_PAZH_MAP                                 = false;
MAP_options.return_log_PAH_MAP                                  = false;
MAP_options.return_p_log_PAZH_p_inv_C_ee_MAP                    = false;
MAP_options.return_p_det_inv_C_aa_p_inv_C_ee_MAP                = false;
MAP_options.return_p_det_inv_C_aa_p_a_MAP                       = false;
MAP_options.return_da_MAP_dinv_C_ee_MAP                         = false;
MAP_options.return_log_ML_MAP                                   = true;  % <-- return the log(Marginal Likelihood), which is calculated using Laplace's method around the MAP point 
MAP_options.return_d_log_ML_MAP_d_inv_C_ee                      = false;
MAP_options.return_log_BFL_MAP                                  = true;  % <-- return the log(Best Fit Likelihood), ditto
MAP_options.return_log_OF_MAP                                   = true;  % <-- return the log(Occam Factor), ditto
% Set options for the optimization
MAP_options.optim_method = 'BFGS_dJ'; % 'Newton' | 'BFGS_dJ' | 'BFGS_dJd2J'
MAP_options.tol          = 1e-10;
MAP_options.verbose      = false; % <-- 'true' outputs the steps during the iteration

% Set starting point for the iteration (this does not have to be a_prior, but a_prior is usually a good choice)
a_init = a_prior;
% Invert C_aa_prior
inv_C_aa_prior = inv(C_aa_prior);
% Inverte C_ee
inv_C_ee = inv(C_ee);

% Call fun_MAP to find the MAP point and ML, BFL, OF using Laplace's method
[a_MAP, ...                         % Parameter values, a, at the MAP point
 ~, ...                             % Measurement weights, inv_C_ee, at the MAP point
 inv_C_aa_MAP, ...                  % Inverse covariance matrix for the parameters, inv_C_aa, at the MAP point
 ~, ...                             % log of normalized posterior likelihood, P(A|Z,H), at the MAP point
 ~, ...                             % log of normalized prior probability P(A|H), at the MAP point
 ~, ...                             % partial(log_PAZH)/partial(inv_C_ee) at the MAP point
 ~, ...                             % partial(det(inv_C_aa))/partial(inv_C_ee) at the MAP point
 ~, ...                             % partial(det(inv_C_aa))/partial(a) at the MAP point
 ~, ...                             % d(a_MAP)/d(inv_C_ee) at the MAP point
 log_ML, ...                        % log of marginal likelihood at the MAP point 
 ~, ...                             % d(log_ML_MAP)/d(inv_C_ee)
 log_BFL, ...                       % log(Best Fit Likelihood)
 log_OF ...                         % log(Occam Factor)
] = fun_MAP(MAP_options,param,a_ref,a_init,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee,z);

% ---------------------------------------------------------------------------
% Print results to screen
% ---------------------------------------------------------------------------
fprintf('------------------------------------------------ \n')
fprintf(' Model is polynomial of order %u                 \n',N_mod)
fprintf('------------------------------------------------ \n')
fprintf('log_ML  = log{ P(D|H) }       = %6.4f \n',log_ML )
fprintf('log_BFL = log{ P(D|A_MAP,H) } = %6.4f \n',log_BFL)
fprintf('log_OF  =                     = %6.4f \n',log_OF )
fprintf('------------------------------------------------ \n')

%--------------------------------------------------------------------------
% Calculate the posterior uncertainties in the model predictions
%--------------------------------------------------------------------------
if plot_options.plot_pred
    % Initialize E and V
    E_y_post = NaN(1,n_plt); V_y_post = NaN(1,n_plt);
    % Loop through the points to calculate the uncertainties in the model predictions
    for n = 1:numel(x_plt)
        param_loc.x = x_plt(n);
        [temp_1,temp_2] = fun_covar_s_P_a(param_loc,a_ref,a_MAP,inv_C_aa_MAP);
        % Extract the expected values from the model
        E_y_post(n) = temp_1;
        % Extract the variance from the model
        V_y_post(n) = temp_2;
    end
    % Plot these
    RGB = gen_RGB;
    % Plot model 1 predictions with errorbars
    han_E = gen_errorpatch(x_plt,E_y_post,V_y_post,RGB.blue);
    plot_options = fun_update_legend(plot_options,han_E,'Model prediction ± 2 s.d.');
end

%--------------------------------------------------------------------------
% Plot to screen or file
%--------------------------------------------------------------------------
if plot_options.plot_gt || plot_options.plot_data || plot_options.plot_pred
    legend(plot_options.plot_handles,plot_options.plot_labels)
    xlabel('x'); ylabel('y'); drawnow
    title('Calculated with P-DATAN (polynomial version of DATAN)')
end

if plot_options.print_to_file
    print('-dpng','Sec_003.png')
end

end

%--------------------------------------------------------------------------
% Utility functions
%--------------------------------------------------------------------------

function [plot_options] = fun_update_legend(plot_options,new_handle,new_label)

% Update the legend handles, and labels
plot_options.plot_handles(end+1) = new_handle;
plot_options.plot_labels{end+1}  = new_label;

end



