function [] = Sec_005()
%--------------------------------------------------------------------------
% Fit several polynomials to data generated from an nth order polynomial with VARIABLE measurement noise 
%--------------------------------------------------------------------------
%
% The assimilation process does not know the ground truth measurement noise, at which the
% simulated data was created. In previous sections, the measurement noise
% for the assimilation was set to the ground truth value. This is
% artificial. In this section, the measurement noise is varied (on a log
% scale) and the data is assimilated at these variable measurement noises. 
%
% As the measurement noise increases (seen in the error bars), the marginal
% likelihood of the lower order polynomials becomes relatively higher. In
% other words, simpler models (lower order polynomials) gain more evidence
% (i.e. are more likely to explain the data) than more complicated models
% (higher order polynomials) as the measurement noise increases. 

%--------------------------------------------------------------------------
% Set plot options.
%--------------------------------------------------------------------------

plot_options.BFL_ML_OF_line = true;
plot_options.model_vs_exp   = true;
plot_options.on_same_fig    = true;
plot_options.print_to_file  = true;

if plot_options.BFL_ML_OF_line || plot_options.model_vs_exp
    RGB = gen_RGB;
    plot_options.colors = {RGB.dark_blue;RGB.blue;RGB.teal;RGB.orange;RGB.dark_orange;RGB.dark_red;RGB.ucam_dark_green;RGB.ucam_dark_purple;RGB.ucam_dark_teal};
end

%--------------------------------------------------------------------------
% Generate ground truth parameters, a_gt, from a polynomial of a user-defined order
%--------------------------------------------------------------------------
% Set the order of the ground truth polynomial
N_gt    = 5;
% Reset the random seed (optional)
rng(1);
% Generate some random data 
x_ref = -0.5:0.05:0.5; n_ref = numel(x_ref); y_ref = randn(n_ref,1);
% Find the coefficients of an order N_gt polynomial that fits this data
a_gt = polyfit(x_ref,y_ref,N_gt);

%--------------------------------------------------------------------------
% Generate N_exp synthetic datapoints from the ground truth polynomial
%--------------------------------------------------------------------------
%
% Set the number of experimental datapoints
N_e  = 10;
% Set the x-locations at which the datapoints are selected
x_e  = linspace(-0.5,0.5,N_e);
% Set the variance of the measurement noise of the ground truth
C_ee = 0.0005;
% Sample the polynomial at the measurement points, corrupted by Gaussian noise 
y_e  = polyval(a_gt,x_e) + sqrt(C_ee)*randn(1,N_e);

%--------------------------------------------------------------------------
% Convert datapoints to P-DATAN format
%--------------------------------------------------------------------------

% Pre-allocate
param = cell(1,N_e); z = cell(1,N_e);
% Put into P-DATAN format
for n_e = 1:N_e
    param{n_e}.x = x_e(n_e);
    z{n_e} = y_e(n_e);
end

%--------------------------------------------------------------------------
% Set the prior and the prior covariances
%--------------------------------------------------------------------------
%
% Set the type of prior
prior = 'zero';
% Set the type of covariance
prior_covariance = 'std_propto_exp_n';

%--------------------------------------------------------------------------
% Set options for fun_MAP
%--------------------------------------------------------------------------
MAP_options.reconverge_to_a_MAP_varying_a                       = true;  % <-- allow parameters to float  
MAP_options.reconverge_to_a_MAP_varying_inv_C_ee                = false; % <-- do not allow measurement noise to float
MAP_options.check_is_MAP_point                                  = false;
MAP_options.force_exact_inv_C_aa_MAP                            = false;
MAP_options.return_inv_C_aa_MAP                                 = false;
MAP_options.return_log_PAZH_MAP                                 = false;
MAP_options.return_log_PAH_MAP                                  = false;
MAP_options.return_p_log_PAZH_p_inv_C_ee_MAP                    = false;
MAP_options.return_p_det_inv_C_aa_p_inv_C_ee_MAP                = false;
MAP_options.return_p_det_inv_C_aa_p_a_MAP                       = false;
MAP_options.return_da_MAP_dinv_C_ee_MAP                         = false;
MAP_options.return_log_ML_MAP                                   = true;  % <-- return the log(Marginal Likelihood), which is calculated using Laplace's method around the MAP point 
MAP_options.return_d_log_ML_MAP_d_inv_C_ee                      = false;
MAP_options.return_log_BFL_MAP                                  = true;  % <-- return the log(Best Fit Likelihood), ditto
MAP_options.return_log_OF_MAP                                   = true;  % <-- return the log(Occam Factor), ditto
% Set options for the optimization
MAP_options.optim_method = 'BFGS_dJ'; % 'Newton' | 'BFGS_dJ' | 'BFGS_dJd2J'
MAP_options.tol          = 1e-10;
MAP_options.verbose      = false; % <-- 'true' outputs the steps during the iteration

%--------------------------------------------------------------------------
% Loop over assumed measurement noise, C_ee, 
%--------------------------------------------------------------------------
%
% Set the number of measurement noises to try
N_ee = 10;
% Create log-distributed C_ee
C_ee_store = logspace(-3,0,N_ee);

for n_ee = 1:N_ee
    % Set C_ee
    C_ee = C_ee_store(n_ee);
    % Normalize C_ee
    [C_ee,C_ee_ref] = gen_fun_normalize(C_ee);

    % Pre-allocate / set to NaN
    log_ML  = NaN(1,9);
    log_BFL = NaN(1,9);
    log_OF  = NaN(1,9);
    a_MAP   = cell(1,9);
    a_ref   = cell(1,9);
    inv_C_aa_MAP = cell(1,9);

    %--------------------------------------------------------------------------
    % Loop over polynomials from 1 to 9
    %--------------------------------------------------------------------------

    for N_mod = 1:9

        % Set the prior
        switch prior
            case 'zero'
                a_prior_unnormalized = zeros(1,N_mod+1);
            otherwise
                disp('prior not set'); return
        end

        % Set the covariance
        switch prior_covariance
            case 'std_propto_exp_n'
                % Approx. match to MC sample of priors
                C_aa_prior_unnormalized = diag(exp(2*(N_mod:-1:0)));
            case 'propto_n'
                % covariance scales with 2^(n), where n is the parameter multiplied by x^n.
                C_aa_prior_unnormalized = 2 * diag(2.^(N_mod:-1:0));
            case 'loose_propto_n'
                C_aa_prior_unnormalized = 2^4 * diag(2.^(N_mod:-1:0));
            case 'very_loose_propto_n'
                C_aa_prior_unnormalized = 2^8 * diag(2.^(N_mod:-1:0));
            case 'unity'
                % covariance is the same for all coefficients
                C_aa_prior_unnormalized = eye(N_mod+1);
            otherwise
                disp('prior covariance not set'); return
        end

        %--------------------------------------------------------------------------
        % Set the reference values for the parameters and noise
        %--------------------------------------------------------------------------
        switch prior_covariance
            case 'std_propto_exp_n'
                a_ref_loc = exp((N_mod:-1:0)).'; % col vector
        end
        % Normalize a_prior and transpose to col vector
        a_prior = a_prior_unnormalized.' ./ a_ref_loc;
        % Normalize C_aa_prior (which should become eye(N+1))
        C_aa_prior = C_aa_prior_unnormalized .* ( (1./a_ref_loc) * (1./a_ref_loc)' );

        %--------------------------------------------------------------------------
        % Calculate the MAP points, the Marginal Likelihood, the BFL and the OF
        %--------------------------------------------------------------------------
        %
        % Set starting point for the iteration (this doesn't have to be a_prior, but it can be)
        a_init = a_prior;
        % Invert C_aa_prior
        inv_C_aa_prior = inv(C_aa_prior);
        % Inverte C_ee
        inv_C_ee = inv(C_ee);

        [a_MAP_, ...                        % Parameter values, a, at the MAP point
            ~, ...                 % Measurement weights, inv_C_ee, at the MAP point
            inv_C_aa_MAP_, ...                 % Inverse covariance matrix for the parameters, inv_C_aa, at the MAP point
            ~, ...                 % log of normalized posterior likelihood, P(A|Z,H), at the MAP point
            ~, ...                  % log of normalized prior probability P(A|H), at the MAP point
            ~, ...    % partial(log_PAZH)/partial(inv_C_ee) at the MAP point
            ~, ...% partial(det(inv_C_aa))/partial(inv_C_ee) at the MAP point
            ~, ...       % partial(det(inv_C_aa))/partial(a) at the MAP point
            ~, ...         % d(a_MAP)/d(inv_C_ee) at the MAP point
            log_ML_, ...                       % log of marginal likelihood at the MAP point
            ~, ...      % d(log_ML_MAP)/d(inv_C_ee)
            log_BFL_, ...                      % log(Best Fit Likelihood)
            log_OF_ ...                        % log(Occam Factor)
            ] = fun_MAP(MAP_options,param,a_ref_loc,a_init,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee,z);

        log_ML(N_mod)  = log_ML_;
        log_BFL(N_mod) = log_BFL_;
        log_OF(N_mod)  = log_OF_;
        a_MAP{N_mod} = a_MAP_;
        a_ref{N_mod} = a_ref_loc;
        inv_C_aa_MAP{N_mod} = inv_C_aa_MAP_;

%         % ---------------------------------------------------------------------------
%         % Print results to screen
%         % ---------------------------------------------------------------------------
%         fprintf('------------------------------------------------ \n')
%         fprintf(' Model is polynomial of order %u                 \n',N_mod)
%         fprintf('------------------------------------------------ \n')
%         fprintf('log_ML  = log{ P(D|H) }       = %6.4f \n',log_ML_ )
%         fprintf('log_BFL = log{ P(D|A_MAP,H) } = %6.4f \n',log_BFL_)
%         fprintf('log_OF  =                     = %6.4f \n',log_OF_ )
%         fprintf('------------------------------------------------ \n')

    end

    %--------------------------------------------------------------------------
    % Plot BFL, ML, and OF to screen
    %--------------------------------------------------------------------------
    if plot_options.BFL_ML_OF_line
        figure(4); clf
        if plot_options.on_same_fig; subplot(3,4,4); else; subplot(3,1,1); end; hold on
        for N_mod = 1:9; plot(N_mod,log_BFL(N_mod),'ks','MarkerFaceColor',plot_options.colors{N_mod},'MarkerSize',12); grid on; end
        title('BFL');
        if plot_options.on_same_fig; subplot(3,4,8); else; subplot(3,1,2); end; hold on
        for N_mod = 1:9; plot(N_mod,log_ML(N_mod),'ks','MarkerFaceColor',plot_options.colors{N_mod},'MarkerSize',12); grid on; end
        title('ML');
        if plot_options.on_same_fig; subplot(3,4,12); else; subplot(3,1,3); end; hold on
        for N_mod = 1:9; plot(N_mod,log_OF(N_mod),'ks','MarkerFaceColor',plot_options.colors{N_mod},'MarkerSize',12); grid on; end
        title('OF'); xlabel('Polynomial order')
    end

    %--------------------------------------------------------------------------
    % Calculate and plot the posterior uncertainties in the model predictions for each model
    %--------------------------------------------------------------------------
    sbplt = [1,2,3,5,6,7,9,10,11];
    if plot_options.model_vs_exp
        if plot_options.on_same_fig; figure(4); else; figure(5); end
        x_plt = -0.5:0.01:0.5; n_plt = numel(x_plt);
        for N_mod = 1:9
            if plot_options.on_same_fig; subplot(3,4,sbplt(N_mod)); hold on; else; subplot(3,3,N_mod); end
            % Initialize E and V
            E_y_post = NaN(1,n_plt); V_y_post = NaN(1,n_plt);
            % Loop through the points to calculate the uncertainties in the model predictions
            for n = 1:numel(x_plt)
                param_loc.x = x_plt(n);
                [temp_1,temp_2] = fun_covar_s_P_a(param_loc,a_ref{N_mod},a_MAP{N_mod},inv_C_aa_MAP{N_mod});
                % Extract the expected values from the model
                E_y_post(n) = temp_1;
                % Extract the variance from the model
                V_y_post(n) = temp_2;
            end
            % Plot model 1 predictions with errorbars
            gen_errorpatch(x_plt,E_y_post,V_y_post,plot_options.colors{N_mod});
            % Plot data with errorbars
            errorbar(x_e,y_e,2*sqrt(C_ee*C_ee_ref)*ones(N_e,1),'k.');
            % Format
            title(sprintf('Order %1u',N_mod)); ylim([-1.5 1.5]); grid on
        end
    end

    drawnow
    if plot_options.print_to_file
        print('-dpng',sprintf('Sec_005_%03u.png',n_ee))
    end

end

end


