function [] = Sec_006()
%--------------------------------------------------------------------------
% Fit several polynomials to data generated from an nth order polynomial with floating measurement noise 
%--------------------------------------------------------------------------
%
% If the measurement noise is reduced, models that miss the datapoints are
% heavily penalized. This heavily penalizes the lower order polynomials. 
%
% In an experiment, we usually under-estimate the measurement error because
% we tend to notice only random error, and not systematic error. Usually we
% are not aware of systematic error because, if we were, we would have
% acted to remove it. 
%
% A way to account for this is to allow the measurement noise to float and to
% simultaneously optimize the parameters AND the measurement noise to
% minimize the negative log likelihood, J. 
%
% This also is a way to incorporate model error. In general, all models
% will contain some model error (e.g. from physics that is missed, or
% measurements that contain systematic error). Here, we account for it in
% the same way that we account for systematic measurement error: by
% increasing the measurement noise. 
%
% fun_MAP in DATAN appends the noise parameters to the model parameters and
% optimizes all simultaneously with gradient-based optimization. This can
% be seen in fun_JdJd2J.m . It is switched on with a flag (see below). 
% 
% This code, Sec_006.m, is otherwise identical to Sec_005.m, so that it is
% easy to see the influence of allowing the measurement noise to float. We
% observe that it is easier to identify the best model (the ground truth
% polynomial in this case). It has a significantly higher marginal likelihood than
% the other models. 

%--------------------------------------------------------------------------
% Generate ground truth parameters, a_gt, from a polynomial of a user-defined order
%--------------------------------------------------------------------------
% Set the order of the ground truth polynomial
N_gt    = 5;
% Reset the random seed (optional)
rng(1);
% Generate some random data 
x_ref = -0.5:0.05:0.5; n_ref = numel(x_ref); y_ref = randn(n_ref,1);
% Find the coefficients of an order N_gt polynomial that fits this data
a_gt = polyfit(x_ref,y_ref,N_gt);

%--------------------------------------------------------------------------
% Generate N_exp synthetic datapoints from the ground truth polynomial
%--------------------------------------------------------------------------
%
% Set the number of experimental datapoints
N_e  = 10;
% Set the x-locations at which the datapoints are selected
x_e  = linspace(-0.5,0.5,N_e);
% Set the variance of the measurement noise
C_ee = 0.0005;
% Sample the polynomial at the measurement points, corrupted by Gaussian noise 
y_e  = polyval(a_gt,x_e) + sqrt(C_ee)*randn(1,N_e);

%--------------------------------------------------------------------------
% Convert datapoints to P-DATAN format
%--------------------------------------------------------------------------
%
% Pre-allocate
param = cell(1,N_e); z = cell(1,N_e);
% Put into P-DATAN format
for n_e = 1:N_e
    param{n_e}.x = x_e(n_e);
    z{n_e} = y_e(n_e);
end

%--------------------------------------------------------------------------
% Set the prior and the prior covariances
%--------------------------------------------------------------------------
% 
% Set the type of prior
prior = 'zero';
% Set the type of covariance
prior_covariance = 'std_propto_exp_n';

%--------------------------------------------------------------------------
% Loop over polynomials from 1 to 9
%--------------------------------------------------------------------------

% Set options for fun_MAP
MAP_options.reconverge_to_a_MAP_varying_a                       = true;    
MAP_options.reconverge_to_a_MAP_varying_inv_C_ee                = true; % <-- allow measurement noise to float
MAP_options.check_is_MAP_point                                  = false;
MAP_options.force_exact_inv_C_aa_MAP                            = false;
MAP_options.return_inv_C_aa_MAP                                 = false;
MAP_options.return_log_PAZH_MAP                                 = false;
MAP_options.return_log_PAH_MAP                                  = false;
MAP_options.return_p_log_PAZH_p_inv_C_ee_MAP                    = false;
MAP_options.return_p_det_inv_C_aa_p_inv_C_ee_MAP                = false;
MAP_options.return_p_det_inv_C_aa_p_a_MAP                       = false;
MAP_options.return_da_MAP_dinv_C_ee_MAP                         = false;
MAP_options.return_log_ML_MAP                                   = true;
MAP_options.return_d_log_ML_MAP_d_inv_C_ee                      = false;
MAP_options.return_log_BFL_MAP                                  = true;
MAP_options.return_log_OF_MAP                                   = true;

MAP_options.optim_method = 'BFGS_dJ'; % 'Newton' | 'BFGS_dJ' | 'BFGS_dJd2J'
MAP_options.tol          = 1e-10;
MAP_options.verbose      = false;

% Normalize C_ee
[C_ee,C_ee_ref] = gen_fun_normalize(C_ee);

% Pre-allocate log_ML, log_BFL, log_OF
log_ML  = NaN(1,9);
log_BFL = NaN(1,9);
log_OF  = NaN(1,9);

for N_mod = 1:9

    % Set the prior
    switch prior
        case 'zero'
            a_prior_unnormalized = zeros(1,N_mod+1);
        otherwise
            disp('prior not set'); return
    end

    % Set the covariance
    switch prior_covariance
        case 'std_propto_exp_n'
            % Approx. match to MC sample of priors
            C_aa_prior_unnormalized = diag(exp(2*(N_mod:-1:0)));
        case 'propto_n'
            % covariance scales with 2^(n), where n is the parameter multiplied by x^n.
            C_aa_prior_unnormalized = 2 * diag(2.^(N_mod:-1:0));
        case 'loose_propto_n'
            C_aa_prior_unnormalized = 2^4 * diag(2.^(N_mod:-1:0));
        case 'very_loose_propto_n'
            C_aa_prior_unnormalized = 2^8 * diag(2.^(N_mod:-1:0));
        case 'unity'
            % covariance is the same for all coefficients
            C_aa_prior_unnormalized = eye(N_mod+1);
        otherwise
            disp('prior covariance not set'); return
    end

    %--------------------------------------------------------------------------
    % Set the reference values for the parameters and noise
    %--------------------------------------------------------------------------
    switch prior_covariance
        case 'std_propto_exp_n'
            a_ref = exp((N_mod:-1:0)).'; % col vector
    end
    % Normalize a_prior and transpose to col vector
    a_prior = a_prior_unnormalized.' ./ a_ref;
    % Normalize C_aa_prior (which should become eye(N+1))
    C_aa_prior = C_aa_prior_unnormalized .* ( (1./a_ref) * (1./a_ref)' );

    %--------------------------------------------------------------------------
    % Calculate the MAP points, the Marginal Likelihood, the BFL and the OF
    %--------------------------------------------------------------------------
    %
    % Set starting point for the iteration (this doesn't have to be a_prior, but it can be)
    a_init = a_prior;
    % Invert C_aa_prior
    inv_C_aa_prior = inv(C_aa_prior);
    % Inverte C_ee
    inv_C_ee = inv(C_ee);

    [~, ...                        % Parameter values, a, at the MAP point
        ~, ...                 % Measurement weights, inv_C_ee, at the MAP point
        ~, ...                 % Inverse covariance matrix for the parameters, inv_C_aa, at the MAP point
        ~, ...                 % log of normalized posterior likelihood, P(A|Z,H), at the MAP point
        ~, ...                  % log of normalized prior probability P(A|H), at the MAP point
        ~, ...    % partial(log_PAZH)/partial(inv_C_ee) at the MAP point
        ~, ...% partial(det(inv_C_aa))/partial(inv_C_ee) at the MAP point
        ~, ...       % partial(det(inv_C_aa))/partial(a) at the MAP point
        ~, ...         % d(a_MAP)/d(inv_C_ee) at the MAP point
        log_ML_, ...                       % log of marginal likelihood at the MAP point
        ~, ...      % d(log_ML_MAP)/d(inv_C_ee)
        log_BFL_, ...                      % log(Best Fit Likelihood)
        log_OF_ ...                        % log(Occam Factor)
        ] = fun_MAP(MAP_options,param,a_ref,a_init,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee,z);

    log_ML(N_mod)  = log_ML_;
    log_BFL(N_mod) = log_BFL_;
    log_OF(N_mod)  = log_OF_;

    % ---------------------------------------------------------------------------
    % Print results to screen
    % ---------------------------------------------------------------------------
    fprintf('------------------------------------------------ \n')
    fprintf(' Model is polynomial of order %u                 \n',N_mod)
    fprintf('------------------------------------------------ \n')
    fprintf('log_ML  = log{ P(D|H) }       = %6.4f \n',log_ML_ )
    fprintf('log_BFL = log{ P(D|A_MAP,H) } = %6.4f \n',log_BFL_)
    fprintf('log_OF  =                     = %6.4f \n',log_OF_ )
    fprintf('------------------------------------------------ \n')

end

%--------------------------------------------------------------------------
% Plot to screen or file
%--------------------------------------------------------------------------
figure(2); clf
subplot(3,1,1); plot(1:9,log_BFL,'o-'); grid on
title('Calculated with P-DATAN with floating noise')
ylabel('BFL'); xlabel('Polynomial order')
subplot(3,1,2); plot(1:9,log_ML,'o-'); grid on
ylabel('ML'); xlabel('Polynomial order')
subplot(3,1,3); plot(1:9,log_OF,'o-'); grid on 
ylabel('OF'); xlabel('Polynomial order')

print('-dpng','Sec_006.png')

end


