function [a_ref,a_polyfit,a_prior,C_aa_prior] = exp_gen_scenario(model,param,z)
% ----------------------------------------------------------------------
% DESCRIPTION
% 
% This function sets the prior variable parameter values, a, for each experimental
% configuration.
%
% For reference, it also calculates the 'ground truth' values, given the
% data. In a real experiment, this is not available, and the experimental
% data would not be supplied to this function. 
%
% ----------------------------------------------------------------------
% INPUTS
%
% model             description of model into which the experimental data is being assimilated  
% param{:}          structure containing the parameters for each experimental configuration: param{1} to param{M}. 
% z{:}              structure containing the experimental data
%
% ----------------------------------------------------------------------
% OUTPUTS
% 
% a_ref             reference values (usually powers of 10)
% a_polyfit         parameter values found with polyfit (for reference)
% a_prior           prior parameter values 
% C_aa_prior        prior covariance matrix (i.e. certainty in the prior parameter values)
%
% Note: for simplicity, a_polyfit is set to 1 and a_ref is set to the ground
% truth value. Normally, a_ref would be a power of 10 and a_polyfit would be
% a number of order 1, which helps the optimization algorithms. 
%
% ----------------------------------------------------------------------

% Set how the prior is calculated
prior = 'close'; % 'actual' | 'zero' | 'close'

switch model
    case '8th_order_polynomial'
        [a_ref,a_polyfit,a_prior,C_aa_prior] = fun_gen_poly(param,z,8,prior);
    case '7th_order_polynomial'
        [a_ref,a_polyfit,a_prior,C_aa_prior] = fun_gen_poly(param,z,7,prior);
    case '6th_order_polynomial'
        [a_ref,a_polyfit,a_prior,C_aa_prior] = fun_gen_poly(param,z,6,prior);
    case '5th_order_polynomial'
        [a_ref,a_polyfit,a_prior,C_aa_prior] = fun_gen_poly(param,z,5,prior);
    case '4th_order_polynomial'
        [a_ref,a_polyfit,a_prior,C_aa_prior] = fun_gen_poly(param,z,4,prior);
    case '3rd_order_polynomial'
        [a_ref,a_polyfit,a_prior,C_aa_prior] = fun_gen_poly(param,z,3,prior);
    case '2nd_order_polynomial'
        [a_ref,a_polyfit,a_prior,C_aa_prior] = fun_gen_poly(param,z,2,prior);
    case '1st_order_polynomial'
        [a_ref,a_polyfit,a_prior,C_aa_prior] = fun_gen_poly(param,z,1,prior);
    case '0th_order_polynomial'
        [a_ref,a_polyfit,a_prior,C_aa_prior] = fun_gen_poly(param,z,0,prior);
    otherwise
        disp('MODEL NOT SPECIFIED IN exp_gen_scenario.m')
end

end

%% Utility functions

function [a_ref,a_polyfit,a_prior,C_aa_prior] = fun_gen_poly(param,z,N,prior)
% param     Configuration
% z         Data
% N         Polynomial order

% Extract the measurement positions
x = cellfun(@(a)getfield(a,'x'),param);
% Extract the experimental data
s = cell2mat(z);
% Find the polynomial coefficients that best fit the data
a_ref = polyfit(x,s,N).';
% Set a_polyfit to 1 so that a_ref * a_polyfit = ground truth 
a_polyfit = ones(N+1,1);
% Set the prior
switch prior
    case 'actual'
        a_prior = a_polyfit;
    case 'zero'
        a_prior  = zeros(N+1,1);
    case 'close'
        a_prior = a_polyfit - 0.01;
    case 'far'
        a_prior = a_polyfit /2;
end
% Set the covariance matrix. For a fair comparison between models, the
% determinants of C_aa_prior must all be the same, so that the Occam
% Factors can be compared.  
C_aa_prior = eye(N+1);

end


