function [J,dJ,d2J,d_det_d2JdAdA_d_inv_C_ee,d_det_d2JdAdA_d_A] = fun_JdJd2J(JdJd2J_options,param,a_ref,a,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee,z)
%
% Calculate J, the negative log likelihood of the posterior distribution of the parameters, A, of the model, H, given data, Z:
%
%           'Posterior'         'Likelihood'         'Prior'
%                |                   |                  |
%  J = - log( P(A|Z,H) ) = - log( P(Z|A,H) ) - log ( P(A|H) )
%                                   |                  |
%                                data Z(A)         parameters A      
%
% INPUTS
%  JdJd2J_options is a structure containing:
%   - what the calculation of J includes
%       include_prior                      (logical)   include P(A|H) in calculation of J  
%       include_z_in_data_likelihood   (logical)   include P(zr|A,H) and P(zi|A,H) in calculation of J 
%   - the derivatives of J that are being requested?
%       calculate_J     (logical)   output J
%       calculate_dJ_   (logical)   output dJ/d(.)
%       calculate_d2J_  (logical)   output d2J/d(.)/d(.)
%   - the variables to which the above derivatives are being requested?
%       by_dA           (logical)   (.) calculate derivatives w.r.t. all the parameters, A
%       by_dinv_C_ee    (logical)   (.) calculate derivatives w.r.t. the (inverse) measurement noise for zr and zi: inv_C_ee
%   - In detail, how should the calculation be performed?
%       normalize_Gaussians     (logical)   Normalize the Likelihood and Prior such that that the integral of P(A|H) is 1 and the integral of P(Z|A,H) for fixed A is 1  
%       d2JdAdA_exact           (logical)   Calculate d2JdAdA exactly (expensive because it uses d2s/dbdb and d2P/dbdb in fun_solve_s) 
%   - whether we need d(det(d2JdAdA))/d(inv_C_ee)
%
%  param{n}        parameter structure (excluding variable parameters, which are given by a_ref * a)
%  a_ref           (real) vector containing reference values of the variable parameters 
%  a               (real) vector containing variable parameters
%  a_prior         (real) vector containing the prior values of the variable parameters 
%  inv_C_aa_prior  (real) prior co-variance matrix of the parameters
%  C_ee_ref        (real) (2x1 col vector) reference values for the measurement noises C_eer and C_eei
%  inv_C_ee        (real) (2x1 col vector) values of the inverse measurement noises, C_eer and C_eei
%  z{n}            (real) experimental measurement
%
% OUTPUTS
%  J                (real scalar)   normalized or un-normalized negative log likelihood
%  dJ               (row vector)    dJ/d(.)      1st derivatives of J w.r.t. requested variables
%  d2J              (matrix)        d2J/d(.)d(.) 2nd derivatives of J w.r.t. requested variables
%
% -------------------------------------------------------------------------

% -------------------------------------------------------------------------
% INTRODUCTION
% -------------------------------------------------------------------------
%
% Terminology of inverse probability (Section 2.3; page 28-32 of David MacKay's book [DM])
% -------------------------------------------------------------------------
%            'Likelihood' 'Prior'
% 'Posterior'      |        |
%     |        P(Z|A,H) * P(A|H)
%  P(A|Z,H)  = -----------------                [1]
%                   P(Z|H)
%                     |
%       'Evidence' / 'Marginal Likelihood'
%
% symbol | description            | label in this script | format
% -------------------------------------------------------------------------
%   H    | the model              | fun_solve_s          | Matlab function
%   A    | the model's parameters | a                    | column vector
%   Z    | the noisy data         | zr, zi, Qr, Qi       | cells of column vectors  
% 
% Likelihood vs Probability (p28/29 of DM)
% -------------------------------------------------------------------------
% P(Z|A,H) is a function of Z and A. It can be thought of in two ways:
%  1. For fixed A, P(Z|A,H) defines a PROBABILITY over Z (which sums to 1 when integrated over all Z)  
%  2. For fixed Z, P(Z|A,H) defines the LIKELIHOOD of A (which does not have to sum to 1 but must be >= 0 and can be much greater than 1; e.g. if the data cannot be explained by the model with any parameter set) 
% In our case, the data Z is fixed but the parameters A are unknown. 
% Therefore P(Z|A,H) is 'the likelihood of the parameters'
%
% Marginal Likelihood (Ex2.6 / p29 / S3.3 / p52 / p298 / p348 of DM)
% -------------------------------------------------------------------------
% P(Z|H) = int P(Z|A,H) P(A|H) dA   is the likelihood of the data, given the
% model after integration over all possible parameter values for that
% model. p298 of DM illustrates why calculation of this integral is usually
% too expensive. 
%
% Maximum Likelihood (Ch22 of DM)
% -------------------------------------------------------------------------
% In the 'Maximum Likelihood method' we find the parameter values, A, that
% maximize the likelihood of A, i.e. we find A for which P(Z|A,H) is
% maximized. This takes no account of the prior P(A|H).
%  Maximum likelihood methods can break down by finding highly-tuned models
% that fit part of the data perfectly. This is known as OVERFITTING. (p306)
% 
% Maximimum a posteriori (MAP) method (p306 of DM)
% -------------------------------------------------------------------------
% This code is written in order to implement the MAP method. We find the
% parameter values, A, that maximize the posterior likelihood of A,
% labelled P(A|Z,H), given fixed data Z, a model, H, and a prior for the
% parameters P(A|H). 
%  In order to find the maximum of P(A|Z,H), we need to maximize the
% numerator of [1]. We do not need to concern ourselves with its
% denominator. In other words ... 
%
%  we find A that maximizes P(A|Z,H) = P(Z|A,H) * P(A|H)
%
% -------------------------------------------------------------------------
% HOW THE MAXIMUM A POSTERIORI METHOD IS REPRESENTED AS J, i.e. A COST FUNCTION FOR OPTIMIZATION 
% -------------------------------------------------------------------------
%
% The likelihood and the prior are modelled as multivariate Gaussian distributions
% -------------------------------------------------------------------------
% Reminder: in the Maximum a posterior method we find A that maximizes
% P(A|Z,H) = P(Z|A,H) * P(A|H). We make the further ASSUMPTIONS that
% P(Z|A,H) and P(A|H) can be modelled as multivariate Gaussian distributions:
%
% 1. P(A|H) 
% 
% Here, the N_a elements of the column vector A are labelled 'a':
%
%           exp{ - 0.5 * (a-a_prior)' * inv_C_aa * (a-a_prior) }    
%  P(A|H) = ----------------------------------------------------    [J1]
%                      sqrt{ (2*pi)^N_a * det(C_aa) }
% 
% where C_aa is the COVARIANCE MATRIX and inv_C_aa is its inverse (a.k.a.
% the PRECISION PARAMETER) (23.2 of DM). This quantifies our certainty in
% our prior estimates for 'a', a_prior. For example, if we do not know any
% of the parameters well beforehand then we would probably set C_aa to be a
% diagonal matrix (i.e. set correlations between the model parameters to
% zero) with large values along the diagonal (i.e. large variances, a.k.a.
% low precision). Note that this is a 'proper prior'. MacKay discourages
% you from using an 'improper prior', such as a uniform distribution. 
%
% 2. P(Z|A,H) 
%
% The types of measurement, Z, are:
%  z        the measurement at x
% 
% The model predictions S(A) are:
%  s(a)     the model prediction at x
%
% We assume that each type of measurement has a noise that is characterized
% by its variance, V. If there are n pings, then the covariance matrix is
% an n x n diagonal matrix with V along the diagonal. (In earlier versions
% of the code, I allowed V to vary from one ping to the next, but this did
% not allow me to optimize for V, which is a useful feature of the code.)
%
% The likelihood, P(Z|A,H), is therefore:
%
%             exp{ - 0.5 * (s(a)-z)' * inv_C_ee * (s(a)-z) }    
%  P(Z|A,H) = ----------------------------------------------   [J2]
%                       sqrt{ (2*pi)^n * det(C_ee) }               
%
%  C_ee    covariance matrix for the measurements
%
% Warning: The Gaussian distribution has very 'light tails' i.e. the
% probabilities go to zero very quickly. MacKay (p312) warns that, while
% single-peaked (unimodal) distributions such as the Gaussian distribution
% are common, most real distributions have heavier tails. 
%
% The negative log likelihood is defined by a cost function J
% -------------------------------------------------------------------------
% We define J to be - log( P(A|Z,H) ), i.e. the negative log of the posterior:  
%
%      <----- contribution from exponential ------>   <- contribution from normalization ->
%  J = 0.5 * (a-a_prior)' * inv_C_aa * (a-a_prior) + log( sqrt{ (2*pi)^N_a * det(C_aa)  } ) [J1] 
%    + 0.5 * (s(a)-z)'    * inv_C_ee * (s(a)-z)    + log( sqrt{ (2*pi)^n   * det(C_ee)  } ) [J2] 
%
% P(A|Z,H) is therefore a maximum at the values of A for which J is a minumum. 
%
% Note that the normalization factors do not depend on A. Therefore they do
% not need to be calculated when finding the minimum of J(A). We only need
% to calculate them when finding the minimum of J w.r.t the measurement
% noise variance, C_ee.
%
% Finding the MAP is therefore the same problem as minimizing J. We will
% minimize J with an optimation algorithm. This will be much quicker if we
% have the 1st derivative of J w.r.t. parameters and might be even quicker
% if we also have the 2nd derivartive of J w.r.t. parameters. 
%
% COMMENT:
% The cost function is formed entirely from real (i.e. not complex) values.
% (There are terminal problems with forming a real cost function from
% complex inputs.)
%
% -------------------------------------------------------------------------
% HOW THE GRADIENT OF J W.R.T. THE PARAMETERS IS CALCULATED
% -------------------------------------------------------------------------
%
% We perform 1st and 2nd order Taylor expansions to obtain dJ/da and d2J/da1/da2  
% -------------------------------------------------------------------------
% During the optimization routine, this function will be fed the column
% vector A = a_l for l in [1,N_a] and be asked to return the following
% real numbers:
%
%  J         scalar
%  dJ/dA     row vector containing dJ/da_l for l in [1,N_a]
%  d2J/dA/dA matrix containing d2J/da_k/da_l for k in [1,N_a] and l in [1,N_a]
%
% Let us consider a perturbation dA to parameter values A0 and approximate
% the new cost function J(A0 + dA) by a 2nd order Taylor expansion around A0:
%
%                           < 1st order >   <------- 2nd order ------->
%  J(A0 + dA)  ~  J(A0)  +  dJ/dA * dA    + 0.5 * dA^T * d2J/dA/dA * dA
%                             |                              |  
%                            row vector                   matrix 
%
% In order to write down dJ/dA and d2J/dA/dA it is convenient to label the
% elements of A with indices k and l, both of which take values in [1,N_a],
% where N_a is the number of parameters. Similarly, the measurements are labelled
% with indices i and j, both of which take values in [1,n], where n is the
% number of measurements. 
%
% REMINDER: The exact expression for J(a) is:
%          <----- contribution from exponential ------>   <- contribution from normalization ->
%  J(a)  = 0.5 * (a-a_prior)' * inv_C_aa  * (a-a_prior) + log( sqrt{ (2*pi)^N_a * det(C_aa_prior) } ) [J1] 
%        + 0.5 * (s(a)-r)'    * inv_C_eer * (s(a)-z)    + log( sqrt{ (2*pi)^n   * det(C_ee)       } ) [J2] 
%  
% For convenience, we drop the contributions from the normalization
% because they do not contain 'a', and we write J(a) in index form: 
%
%  J(a)  = 0.5 * (a-a_prior)_k * inv_C_aa_kl * (a-a_prior)_l  [J1] 
%        + 0.5 * (s(a)-z)_i    * inv_C_ee_ij * (s(a)-z)_j     [J2] 
%
% Now we perform the Taylor expansion:
%
%  J(a0 + da) ~ 0.5 * (a0-a_prior)_k * inv_C_aa_kl  * (a0-a_prior)_l  [J1] 
%             + 0.5 * (s(a0)-z)_i    * inv_C_ee_ij  * (s(a0)-z)_j     [J2] 
%           + {
%               (a0-a_prior)_k * inv_C_aa_kl                          [dJ1/da_l]  }         
%             + (s(a0)-z)_i    * inv_C_ee_ij * ds_j/da_l|a0           [dJ2/da_l]  }  row vector dJ/dA       
%             } * da_l    
%           + 0.5 * da_k * {
%               inv_C_aa_kl                                                                                        [d2J1/da_k/da_l]  }    
%             + ds_i/da_k|a0 * inv_C_ee_ij * ds_j/da_l|a0 + (s(a0)-z)_i  * inv_C_ee_ij * d2s_j/da_k/da_l|a0        [d2J2/da_k/da_l]  }  matrix d2J/dA/dA       
%             } * da_l
%
% This function outputs J, dJ/dA, and d2J/dA/dA
%
% 
% -------------------------------------------------------------------------
% Technical details about how J, dJ/dA, and d2J/dA/dA are formed
% -------------------------------------------------------------------------
% J = J1 (parameter discrepancy) is calculated first
% for each experimental configuration:
%   J = J + J2 (growth rate discrepancy) 
% end
% 
% If the contributions from the normalizations are required, these are
% added within the above loops. 
%
% dJ/dA and d2J/dA/dA are formed with similar loops 
%
% -------------------------------------------------------------------------
% HOW THE GRADIENT OF J W.R.T. THE MEASUREMENT NOISE IS CALCULATED
% -------------------------------------------------------------------------
% fun_SP_a_C_s_P.m finds the measurement noise that maximizes the a
% posteriori likelihood of the parameters. For this, it needs:
%  dJ/dC_ee, d2J/da/dC_ee, and d2J/dCee^2
%
% These are found in the same way as dJ/dA and d2J/dA/dA
%
% -------------------------------------------------------------------------
% TECHNICAL DETAILS
% -------------------------------------------------------------------------
% a is scaled by a_ref such that all elements of a should be of order 1.
% This is good practice when using primitive optimization routines such as
% steepest descent. For sophisticated algorithms such as BFGS it should not
% be necessary but I find that it helps nonetheless. For the Newton method,
% it makes no difference. If elements of 'a' turn out to be far from order 1,
% it is worth renormalizing 'a' with gen_fun_normalize.m. 
%  The procedure is this:
%   - this function sends (a_ref .* a) to fun_solve_s.m
%   - fun_solve_s.m returns d[]db and d2[]dbdb based on (a_ref .* a) 
%   - this function rescales d[]db and d2[]dbdb to be based on a
% Therefore fun_solve_s.m and its lower level functions are
% unaffected by the addition of a_ref. 
%
% Similarly inv_C_** is scaled by C_**_ref, which is a vector containing typical
% reference values for the noise covariances, C_**, which are fed in here as
% inv_C_**. The aim is to keep the optimized values of C_** close to 1.
%  The procedure is
%   - C_**_ref is included when the inverse variance of the measurements are
%   unpacked
%   - this function sends (inv_C_** ./ C_**_ref) to fun_solve_s.m
%   - fun_solve_s.m returns d[]d(C_**) and d2[]/dC_**/dC_**
%   - fun_JdJd2J rescales them. 
%
% The covariance matrices are sent as their inverses (i.e. their
% 'precisions', rather than their 'variances') because this requires fewer
% matrix inversions. 
%
% -------------------------------------------------------------------------
% POSSIBLE FUTURE IMPROVEMENTS
% -------------------------------------------------------------------------
% Storing s for each experiment
%  At the moment s is thrown away on function end. The next evaluation of J
%  therefore starts from param.freq, which may be rather far from s. It may
%  be better to return the evaluated values of s for each experiment and
%  then re-use these during the next function evaluation. 

%% Count the inputs and check for inconsistencies between what has been asked for and what has been provided
% -------------------------------------------------------------------------
% Count the number of configurations (each with their own x)
% -------------------------------------------------------------------------
N_c = length(param);
% -------------------------------------------------------------------------
% Count the number of variable parameters (This is zero if the parameters in 'param' are not to be varied) 
% -------------------------------------------------------------------------
N_a = length(a_ref);
% -------------------------------------------------------------------------
% If the data likelihood is to be calculated, either empty 'varparams' or push the variable parameters into 'varparams' 
% -------------------------------------------------------------------------
if JdJd2J_options.include_z_in_data_likelihood
    if N_a == 0
        % Empty varparams
        param = cellfun(@(cell_n)setfield(cell_n,'varparams',[]),param,'UniformOutput',false);
    else
        % Push the variable parameters (scaled by a_ref) into param.varparams for each configuration
        param = cellfun(@(cell_n)setfield(cell_n,'varparams',a_ref .* a),param,'UniformOutput',false);
    end
end
% -------------------------------------------------------------------------
% Extract the growth rate and frequency measurement noises
% -------------------------------------------------------------------------
if JdJd2J_options.include_z_in_data_likelihood
    % Count the number of gr.rate and frequency measurements per ping (hard-wired to 1)
%     N_e = 1;
end
% -------------------------------------------------------------------------
% Set the number of gradients w.r.t. parameters and measurement noises
% -------------------------------------------------------------------------
if or(JdJd2J_options.calculate_dJ_,JdJd2J_options.calculate_d2J_)
    % -------------------------------------------------------------------------
    % Set the number of gradients w.r.t. parameters that have been requested
    % -------------------------------------------------------------------------
    if JdJd2J_options.by_dA
        if JdJd2J_options.include_prior
            N_aa = N_a;
        else
            fprintf('WARNING!!: You have asked for gradients w.r.t. the parameters without providing the parameters \n\n\n')
            return
        end
    else
        N_aa = 0;
    end
    % -------------------------------------------------------------------------
    % Set the number of gradients w.r.t. gr.rate/frequency noises there are
    % -------------------------------------------------------------------------
    if JdJd2J_options.by_dinv_C_ee
        if JdJd2J_options.include_z_in_data_likelihood
%             N_ee = N_e;
        else
            fprintf('WARNING!! You have asked for gradients w.r.t. inv_C_ee without providing zr and zi \n\n\n')
            return
        end
    else
%         N_ee = 0;
    end
end

%% Initialize the variables that will be compiled into the output variables
% -------------------------------------------------------------------------
% initialize J
% -------------------------------------------------------------------------
if JdJd2J_options.calculate_J
    J = 0;
else
    J = [];
end
% -------------------------------------------------------------------------
% initialize dJ/d(various)
% -------------------------------------------------------------------------
if JdJd2J_options.calculate_dJ_
    % ---------------------------------------------------------------------
    % initialize dJ/dA
    % ---------------------------------------------------------------------
    if JdJd2J_options.by_dA
        dJdA = zeros(1,N_aa);
    else
        dJdA = [];
    end
    % ---------------------------------------------------------------------
    % initialize dJ/d(inv_C_ee)
    % ---------------------------------------------------------------------
    if JdJd2J_options.by_dinv_C_ee
        dJd_inv_C_ee = 0; 
    else
        dJd_inv_C_ee = []; 
    end
end
% -------------------------------------------------------------------------
% initialize d2J/d(various)/d(various)
% -------------------------------------------------------------------------
if JdJd2J_options.calculate_d2J_
    % ---------------------------------------------------------------------
    % initialize d2J/dA/dA
    % ---------------------------------------------------------------------
    if JdJd2J_options.by_dA
        d2JdAdA = zeros(N_aa);
    else
        d2JdAdA = [];
    end
    % ---------------------------------------------------------------------
    % initialize d2J/d(A)/d(inv_C_ee)
    % ---------------------------------------------------------------------
    if JdJd2J_options.by_dA && JdJd2J_options.by_dinv_C_ee
        d2JdAd_inv_C_ee = zeros(1,N_aa);
    else
        d2JdAd_inv_C_ee = [];
    end
    % ---------------------------------------------------------------------
    % initialize d2J/d(inv_C_ee)/d(inv_C_ee)
    % ---------------------------------------------------------------------
    if JdJd2J_options.by_dinv_C_ee
        d2Jd_inv_C_ee_d_inv_C_ee = 0; 
    else
        d2Jd_inv_C_ee_d_inv_C_ee = []; 
    end
end
% -------------------------------------------------------------------------
% initialize d3J/d(A)/d(A)/d(inv_C_ee), which is for d(det(d2JdAdA))/d(inv_C_ee)
% -------------------------------------------------------------------------
if JdJd2J_options.d_det_d2JdAdA_d_inv_C_ee
    if JdJd2J_options.calculate_d2J_
        % ---------------------------------------------------------------------
        % initialize d3J/dA/dA/d(inv_C_ee)
        % ---------------------------------------------------------------------
        if JdJd2J_options.by_dA
            d3JdAdAd_inv_C_ee = zeros(N_aa);
        else
            d3JdAdAd_inv_C_ee = [];
        end
    else
        disp('WARNING: You have asked for the gradient of d2JdAdA but have not asked to calculate d2JdAdA')
    end
end
% -------------------------------------------------------------------------
% initialize d3J/d(A)/d(A)/d(A), which is for d(det(d2JdAdA))/d(A)
% -------------------------------------------------------------------------
if JdJd2J_options.d_det_d2JdAdA_dA
    if JdJd2J_options.calculate_d2J_
        % ---------------------------------------------------------------------
        % initialize d3J/dA/dA/d(A)
        % ---------------------------------------------------------------------
        if JdJd2J_options.by_dA
            d3JdAdAdA = zeros(N_aa,N_aa,N_aa);
        else
            d3JdAdAdA = [];
        end
    else
        disp('WARNING: You have asked for the gradient of d2JdAdA but have not asked to calculate d2JdAdA')
    end
end

%% Calculate J1 (the prior) and all its requested derivatives
%
%          <----- contribution from exponential ------>   <- contribution from normalization ->
%  J(a)  = 0.5 * (a-a_prior)' * inv_C_aa  * (a-a_prior) + log( sqrt{ (2*pi)^N_a * det(C_aa_prior)  } ) [J1] 
%
if JdJd2J_options.include_prior
    % ---------------------------------------------------------------------
    % Calculate J1
    % ---------------------------------------------------------------------
    if JdJd2J_options.calculate_J
        % Contribution from the exponential
        J = J + 0.5 * (a - a_prior).' * inv_C_aa_prior * (a - a_prior);
        if JdJd2J_options.normalize_Gaussians
            % Contribution from the normalization
            J = J + 0.5 * N_a * log(2*pi) - sum(log(diag(chol(inv_C_aa_prior))));
        end
    end
    % ---------------------------------------------------------------------
    % Calculate d(J1)/d(A)
    % ---------------------------------------------------------------------
    if JdJd2J_options.calculate_dJ_
        if JdJd2J_options.by_dA
            dJdA = dJdA + (a - a_prior).' * inv_C_aa_prior ;
        end
    end
    % ---------------------------------------------------------------------
    % Calculate d2(J1)/d(A)/d(A)
    % ---------------------------------------------------------------------
    if JdJd2J_options.calculate_d2J_
        if JdJd2J_options.by_dA
            d2JdAdA = d2JdAdA + inv_C_aa_prior;
        end
    end
end

%% Cycle through the experiments to calculate the contributions to J, dJ, and d2J from the experimental measurement discrepancy
%
for nn = 1:N_c
    % ---------------------------------------------------------------------
    % Calculate the model predictions s(a) and P(a) for configuration nn
    % ---------------------------------------------------------------------
    if JdJd2J_options.include_z_in_data_likelihood
        % Calculate s for experiment nn
        s = fun_solve_s(param{nn});
        % Count the number of measurements (pings) in experiment nn
        N_p = length(z{nn});
    end
    % ---------------------------------------------------------------------
    % Calculate the contribution to J from configuration nn
    % ---------------------------------------------------------------------
    %          <----- contribution from exponential ------>   <- contribution from normalization ->
    %  J(a)  = 0.5 * (sr(a)-zr)'  * inv_C_eer * (sr(a)-zr)  + log( sqrt{ (2*pi)^n   * det(C_eer) } ) [J2]
    if JdJd2J_options.calculate_J 
        if JdJd2J_options.include_z_in_data_likelihood 
            % Contribution from the exponential terms in J2
            J = J + 0.5 * (s*ones(N_p,1) - z{nn}).' * (s*ones(N_p,1) - z{nn}) * inv_C_ee/C_ee_ref; ...
            if JdJd2J_options.normalize_Gaussians
                % Contribution from the normalization terms in J2
                J = J + 0.5 * N_p * ( log(2*pi) - log(inv_C_ee/C_ee_ref) ); 
            end
        end
    end
    % ---------------------------------------------------------------------
    % Calculate the contribution to d(J2)/d(inv_C_ee)
    % ---------------------------------------------------------------------
    if JdJd2J_options.calculate_dJ_
        if JdJd2J_options.by_dinv_C_ee
            if JdJd2J_options.include_z_in_data_likelihood
                % Contribution from the exponential terms in J2
                dJd_inv_C_ee = dJd_inv_C_ee + 0.5 * (s*ones(N_p,1) - z{nn}).' * (s*ones(N_p,1) - z{nn}) / C_ee_ref;
                if JdJd2J_options.normalize_Gaussians
                    % Contribution from the normalization terms in J2
                    dJd_inv_C_ee = dJd_inv_C_ee - 0.5 * N_p/inv_C_ee ;
                end
            else
                fprint('WARNING!! dJ/d(inv_C_ee) requested but zr and zi not provided\n\n\n')
            end
        end
    end
    % ---------------------------------------------------------------------
    % Calculate the model predictions ds/da and dP/da for configuration nn
    % ---------------------------------------------------------------------
    % The following cases require dsrdb and dsidb:
    %     if assim_options.dJ_
    %         if assim_options.by_dA
    %             if assim_options.include_zr_zi_in_data_likelihood
    %     if assim_options.d2J_
    %         if assim_options.by_dA
    %             if assim_options.by_dinv_C_ee
    %                 if assim_options.include_zr_zi_in_data_likelihood
    %     if assim_options.d2J_
    %         if assim_options.by_dA
    %             if assim_options.include_zr_zi_in_data_likelihood
    %
    % The following cases require dPrdb and dPidb:
    %     if assim_options.dJ_
    %         if assim_options.by_dA
    %             if assim_options.include_Qr_Qi_in_data_likelihood
    %     if assim_options.d2J_
    %         if assim_options.by_dA
    %             if assim_options.by_dinv_C_pp
    %                 if assim_options.include_Qr_Qi_in_data_likelihood
    %     if assim_options.d2J_
    %         if assim_options.by_dA
    %             if assim_options.include_Qr_Qi_in_data_likelihood
    if or(JdJd2J_options.calculate_dJ_,JdJd2J_options.calculate_d2J_)
        if JdJd2J_options.by_dA
            if JdJd2J_options.include_z_in_data_likelihood
                % Calculate dsdb for this experiment
                [~,dsdb] = fun_solve_s(param{nn});
                % Rescale by a_ref
                dsdb = dsdb .* a_ref;
            end
        end
    end
    % ---------------------------------------------------------------------
    % Calculate the contribution to dJdA from configuration nn
    % ---------------------------------------------------------------------
    %             + (s(a0)-z)_i  * inv_C_ee_ij * ds_j/da_l|a0         [dJ2/da_l]  } row vector dJ/dA 
    if JdJd2J_options.calculate_dJ_
        if JdJd2J_options.by_dA
            if JdJd2J_options.include_z_in_data_likelihood
                % Contribution from the exponential terms in J2
                dJdA = dJdA ...
                    + (s*N_p - sum(z{nn})) * dsdb.' * inv_C_ee/C_ee_ref;
            end
        end
    end
    % ---------------------------------------------------------------------
    % Calculate the contribution to d2JdAd(inv_C_ee*) from configuration nn
    % ---------------------------------------------------------------------
    if JdJd2J_options.calculate_d2J_
        if JdJd2J_options.by_dA
            if JdJd2J_options.by_dinv_C_ee
                if JdJd2J_options.include_z_in_data_likelihood
                    % Contribution from the exponential terms in J2
                    d2JdAd_inv_C_ee = d2JdAd_inv_C_ee + (s*N_p - sum(z{nn})) * dsdb.' / C_ee_ref;
                    % The normalization terms in J2 do not contribute
                end
            end
        end
    end
    % ---------------------------------------------------------------------
    % Calculate the contribution to d2Jd(inv_C_ee*)d(inv_C_ee*) from experiment nn
    % ---------------------------------------------------------------------
    if JdJd2J_options.calculate_d2J_
        if JdJd2J_options.by_dinv_C_ee
            if JdJd2J_options.include_z_in_data_likelihood
                % Contribution from the normalization terms in J2
                if JdJd2J_options.normalize_Gaussians
                    d2Jd_inv_C_ee_d_inv_C_ee = d2Jd_inv_C_ee_d_inv_C_ee + 0.5 * N_p/inv_C_ee^2 ;
                end
                % There is no contribution from the exponential term in J2
            end
        end
    end
    % ---------------------------------------------------------------------
    % Calculate the contribution to d2JdAdA from configuration nn
    % ---------------------------------------------------------------------
    %             + ds_i/da_k|a0 * inv_C_ee_ij * ds_j/da_l|a0 + (s(a0)-z)_i  * inv_C_ee_ij * d2s_j/da_k/da_l|a0        [d2J2/da_k/da_l]   } matrix d2J/dA/dA 
    %               <--------------- cheap term ------------>   <-------------- expensive term --------------->
    %
    % NOTE
    %  There are three ways to calculate d2JdAdA:
    %    1. Ignoring the expensive term (Gauss-Newton) 
    %    2. Including the expensive term (exact but slow)
    %    3. Estimating d2JdAdA during the BFGS_dJ algorithm (assim_options.approx_Hessian = true)
    %  Method 3 is quick, cheap, and seems to be nearly as good as 2
    %
    if JdJd2J_options.calculate_d2J_
        if JdJd2J_options.by_dA
            % -----------------------------------------------------------------
            % Calculate the contribution to d2JdAdA from the cheap term
            % -----------------------------------------------------------------
            if JdJd2J_options.include_z_in_data_likelihood
                % Contribution from the exponential terms in J2
                d2JdAdA = d2JdAdA ...
                    + N_p * (dsdb * dsdb.') * inv_C_ee/C_ee_ref;
                % There is no contribution from the normalization term in J2
                if JdJd2J_options.d_det_d2JdAdA_d_inv_C_ee
                    d3JdAdAd_inv_C_ee = d3JdAdAd_inv_C_ee ...
                    + N_p * (dsdb * dsdb.') *       1/C_ee_ref;
                end
                % *********************************************************
                % NEW TERMS; NOT YET SURE THESE ARE CORRECT or whether they can be calculated more simply 
                % (For the polynomial, they are always zero anyway)
                if JdJd2J_options.d_det_d2JdAdA_dA
                    % Calculate d2sdbdb for this experiment
                    [~,~,d2sdbdb] = fun_solve_s(param{nn}); % MOVE THIS; I AM CALCULATING THIS TWICE
                    % Rescale by a_ref
                    d2sdbdb = d2sdbdb .* (a_ref * a_ref.');
                    % Loop through a, calculating d(d2JdAdA)/d(A)
                    for n_a = 1:N_a
                        d3JdAdAdA(:,:,n_a) = d3JdAdAdA(:,:,n_a) ...
                            + N_p * (d2sdbdb(:,n_a) * dsdb.' + dsdb * d2sdbdb(:,n_a).') * inv_C_ee/C_ee_ref;
                    end
                end
                % END OF NEW TERMS
                % *********************************************************
            end
            if JdJd2J_options.d2JdAdA_exact
                % -----------------------------------------------------------------
                % Calculate the contribution to d2JdAdA from the expensive term
                % -----------------------------------------------------------------
                % ---------------------------------------------------------------------
                % Calculate the model predictions d2s/da/da and d2P/da/da for configuration nn
                % ---------------------------------------------------------------------
                if JdJd2J_options.include_z_in_data_likelihood
                    % Calculate d2sdbdb for this experiment
                    [~,~,d2sdbdb] = fun_solve_s(param{nn});
                    % Rescale by a_ref
                    d2sdbdb = d2sdbdb .* (a_ref * a_ref.');
                end
                % -----------------------------------------------------------------
                % Calculate the contribution to d2JdAdA from the expensive term
                % -----------------------------------------------------------------
                if JdJd2J_options.include_z_in_data_likelihood
                    % Contribution from the exponential terms in J2 and J3
                    d2JdAdA = d2JdAdA ...
                        + (s*N_p - sum(z{nn})) * d2sdbdb * inv_C_ee/C_ee_ref;
                    % There is no contribution from the normalization term
                    if JdJd2J_options.d_det_d2JdAdA_d_inv_C_ee
                        d3JdAdAd_inv_C_ee = d3JdAdAd_inv_C_ee ...
                        + (s*N_p - sum(z{nn})) * d2sdbdb *        1/C_ee_ref;
                    end
                    % *********************************************************
                    % NEW TERMS; NOT YET SURE THESE ARE CORRECT or whether they can be calculated more simply
                    % (For the polynomial, they are always zero anyway)
                    if JdJd2J_options.d_det_d2JdAdA_dA
                        % Loop through n_a
                        for n_a = 1:N_a
                            d3JdAdAdA(:,:,n_a) = d3JdAdAdA(:,:,n_a) ...
                                + (dsdb(n_a)*N_p) * d2sdbdb * inv_C_ee/C_ee_ref;
                            % (I neglect the d3sdbdb term; it is too hard to calculate)
                        end
                    end
                    % END OF NEW TERMS
                    % *********************************************************
                end
            end
        end
    end
end

% -----------------------------------------------------------------
% Assemble dJ
% -----------------------------------------------------------------
if JdJd2J_options.calculate_dJ_
    % Compile the row vector of first derivatives
    dJ =  [dJdA             , dJd_inv_C_ee];
else
    dJ = [];
end

% -----------------------------------------------------------------
% Assemble d2J
% -----------------------------------------------------------------
if JdJd2J_options.calculate_d2J_
    % Compile the matrix of second derivatives
    d2J = [d2JdAdA          , d2JdAd_inv_C_ee.'        ; ...
           d2JdAd_inv_C_ee  , d2Jd_inv_C_ee_d_inv_C_ee ] ;
else
    d2J = [];
end

% -----------------------------------------------------------------
% Assemble d(det(d2JdAdA))/d(inv_C_ee)
% -----------------------------------------------------------------
% !!! THIS ALWAYS EQUALS THE NUMBER OF MEASUREMENT POINTS !!!
% (This may be very very obvious but it may also be an artefact of this
% simple case. I will leave the full calculation in for now.)
if JdJd2J_options.d_det_d2JdAdA_d_inv_C_ee
    % Calculate d_det_d2JdAdA_d_inv_C_ee with Jacobi's formula
    d_det_d2JdAdA_d_inv_C_ee = trace(adjoint(d2JdAdA) * d3JdAdAd_inv_C_ee);
else
    d_det_d2JdAdA_d_inv_C_ee = [];
end

% -----------------------------------------------------------------
% Assemble d(det(d2JdAdA))/d(A)
% -----------------------------------------------------------------
if JdJd2J_options.d_det_d2JdAdA_dA % d3JdAdAdA
    % Calculate d_det_d2JdAdA_d_A with Jacobi's formula
    d_det_d2JdAdA_d_A = NaN(N_a,1);
    for n_a = 1:N_a
        d_det_d2JdAdA_d_A(n_a) = trace(adjoint(d2JdAdA) * d3JdAdAdA(:,:,n_a));
    end
else
    d_det_d2JdAdA_d_A = [];
end

end
