function [a_MAP, ...                        % Parameter values, a, at the MAP point
          inv_C_ee_MAP, ...                 % Measurement weights, inv_C_ee, at the MAP point
          inv_C_aa_MAP, ...                 % Inverse covariance matrix for the parameters, inv_C_aa, at the MAP point
          log_PAZH_MAP, ...                 % log of normalized posterior likelihood, P(A|Z,H), at the MAP point
          log_PAH_MAP, ...                  % log of normalized prior probability P(A|H), at the MAP point
          p_log_PAZH_p_inv_C_ee_MAP, ...    % partial(log_PAZH)/partial(inv_C_ee) at the MAP point
          p_det_inv_C_aa_p_inv_C_ee_MAP, ...% partial(det(inv_C_aa))/partial(inv_C_ee) at the MAP point
          p_det_inv_C_aa_p_a_MAP, ...       % partial(det(inv_C_aa))/partial(a) at the MAP point
          da_MAP_dinv_C_ee_MAP, ...         % d(a_MAP)/d(inv_C_ee) at the MAP point
          log_ML_MAP, ...                   % log of marginal likelihood at the MAP point 
          d_log_ML_MAP_d_inv_C_ee, ...      % d(log_ML_MAP)/d(inv_C_ee)
          log_BFL, ...                      % log(Best Fit Likelihood)
          log_OF ...                        % log(Occam Factor)
          ] = fun_MAP(MAP_options,param,a_ref,a_init,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee_init,z)

% CUT, PASTE, AND UNCOMMENT THE TEXT BELOW TO CALL THIS FUNCTION
% MAP_options.reconverge_to_a_MAP_varying_a                       = true;    
% MAP_options.reconverge_to_a_MAP_varying_inv_C_ee                = false;
% MAP_options.check_is_MAP_point                                  = false;
% MAP_options.force_exact_inv_C_aa_MAP                            = false;
% MAP_options.return_inv_C_aa_MAP                                 = false;
% MAP_options.return_log_PAZH_MAP                                 = false;
% MAP_options.return_log_PAH_MAP                                  = false;
% MAP_options.return_p_log_PAZH_p_inv_C_ee_MAP                    = false;
% MAP_options.return_p_det_inv_C_aa_p_inv_C_ee_MAP                = false;
% MAP_options.return_p_det_inv_C_aa_p_a_MAP                       = false;
% MAP_options.return_da_MAP_dinv_C_ee_MAP                         = false;
% MAP_options.return_log_ML_MAP                                   = false;
% MAP_options.return_d_log_ML_MAP_d_inv_C_ee                      = false;
% MAP_options.return_log_BFL_MAP                                  = false;
% MAP_options.return_log_OF_MAP                                   = false;
% MAP_options.optim_method = 'BFGS_dJ'; % 'Newton' | 'BFGS_dJ' | 'BFGS_dJd2J'
% MAP_options.tol          = 1e-10;
% MAP_options.verbose      = false;
%
% [a_MAP, ...                        % Parameter values, a, at the MAP point
%  inv_C_ee_MAP, ...                 % Measurement weights, inv_C_ee, at the MAP point
%  inv_C_aa_MAP, ...                 % Inverse covariance matrix for the parameters, inv_C_aa, at the MAP point
%  log_PAZH_MAP, ...                 % log of normalized posterior likelihood, P(A|Z,H), at the MAP point
%  log_PAH_MAP, ...                  % log of normalized prior probability P(A|H), at the MAP point
%  p_log_PAZH_p_inv_C_ee_MAP, ...    % partial(log_PAZH)/partial(inv_C_ee) at the MAP point
%  p_det_inv_C_aa_p_inv_C_ee_MAP, ...% partial(det(inv_C_aa))/partial(inv_C_ee) at the MAP point
%  p_det_inv_C_aa_p_a_MAP, ...       % partial(det(inv_C_aa))/partial(a) at the MAP point
%  da_MAP_dinv_C_ee_MAP, ...         % d(a_MAP)/d(inv_C_ee) at the MAP point
%  log_ML_MAP, ...                   % log of marginal likelihood at the MAP point 
%  d_log_ML_MAP_d_inv_C_ee, ...      % d(log_ML_MAP)/d(inv_C_ee)
%  log_BFL, ...                      % log(Best Fit Likelihood)
%  log_OF ...                        % log(Occam Factor)
% ] = fun_MAP(MAP_options,param,a_ref,a_init,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee_init,z)

% -------------------------------------------------------------------------
% SUMMARY
% -------------------------------------------------------------------------
% The function fun_JdJd2J.m calculates and returns J, which is the
% negative log posterior likelihood of the parameters given the data:
%
%           'Posterior'         'Likelihood'         'Prior'
%                |                   |                  |
%  J = - log( P(A|Z,H) ) = - log( P(Z|A,H) ) - log ( P(A|H) )
%
% J contains contributions from the prior [J1] and the measurements [J2]:
%      <----- contribution from exponential ----->   <- contribution from normalization ->
%  J = 0.5 * (a-a_prior)' * inv_C_aa * (a-a_prior) + log( sqrt{ (2*pi)^N_a * det(C_aa)  } ) [J1] 
%    + 0.5 * (sr(a)-zr)'  * inv_C_ee * (sr(a)-zr)  + log( sqrt{ (2*pi)^n   * det(C_ee)  } ) [J2] 
%
% where the matrices (sent as their inverses) are:
%  C_aa    covariance matrix for the prior
%  C_ee    covariance matrix for the measurements
% 
% -------------------------------------------------------------------------
% MAXIMIZING THE A-POSTERIORI LIKELIHOOD P(A|Z,H)
% -------------------------------------------------------------------------
% Minimizing J corresponds to maximizing the a-posterior likelihood,
% P(A|Z,H). If asked, this function finds the values of a, C_ee, or both at
% which P(A|Z,H) is maximized. This is called the MAP point.
%
% -------------------------------------------------------------------------
% CALCULATING THE MARGINAL LIKELIHOOD
% -------------------------------------------------------------------------
% The Marginal Likelihood, P(Z|H), can be calculated with Laplace's method:
%
%   Posterior likelihood: P(A_MAP|Z,H) including normalizations
%                            |
%           <----------------------->
%  P(Z|H) ~ P(Z|A_MAP,H) * P(A_MAP|H) * 1/sqrt(det(inv_C_aa_MAP/2pi))   [1]
%     |            |       <---------------------------------------->
%     |            |                           |
%  Evidence  Best fit likelihood          Occam factor
%
% -------------------------------------------------------------------------
% CALCULATING GRADIENTS
% This function also calculates the gradients of useful values with respect
% to the inputs. See the comments below for details. 
% 
% -------------------------------------------------------------------------
% NOTES:
%  The aim is that this is the ONLY function that calls fun_JdJd2J and is
%  also the only function that you need to assimilate the MAP point. Another
%  function wraps around this one to maximize the marginal likelihood by
%  changing the noise. 

%% ========================================================================
% Map from MAP_options to int_options
% =========================================================================
%
% Note: This seems trivial but it makes the code easier to read.
%
% -------------------------------------------------------------------------
% Set all int_options to false
% -------------------------------------------------------------------------
int_options.reconverge                                          = false;
int_options.calc_logPAZH_MAP                                    = false;
int_options.calc_logPAH_MAP                                     = false;
int_options.calc_inv_C_aa_MAP                                   = false;
int_options.calc_log_ML_MAP                                     = false;
int_options.calc_p_log_PAZH_p_inv_C_ee_MAP                      = false;
int_options.calc_p_det_inv_C_aa_p_inv_C_ee_MAP                  = false;
int_options.calc_p_det_inv_C_aa_p_a_MAP                         = false;
int_options.calc_d_log_ML_MAP_d_inv_C_ee                        = false;
int_options.calc_da_MAP_dinv_C_ee_MAP                           = false;
int_options.calc_log_PAH_MAP                                    = false;
% -------------------------------------------------------------------------
% Reconverge to a_MAP if either of these is true:
% - MAP_options.reconverge_to_a_MAP_varying_a
% - MAP_options.reconverge_to_a_MAP_varying_inv_C_ee
% If not reconverging, check that a_init is a MAP point, if asked
% -------------------------------------------------------------------------
if or(MAP_options.reconverge_to_a_MAP_varying_a,MAP_options.reconverge_to_a_MAP_varying_inv_C_ee)
    int_options.reconverge                                      = true;
else
    int_options.check_is_MAP_point = MAP_options.check_is_MAP_point;
end
% -------------------------------------------------------------------------
% Return inv_C_aa at the MAP point
% -------------------------------------------------------------------------
if MAP_options.return_inv_C_aa_MAP
    int_options.calc_inv_C_aa_MAP                               = true;
end
% -------------------------------------------------------------------------
% Return log(P(A|Z,H)) at the MAP point
% -------------------------------------------------------------------------
if MAP_options.return_log_PAZH_MAP
    int_options.calc_logPAZH_MAP                                = true;
end
% -------------------------------------------------------------------------
% Return log(P(A|H)) at the MAP point
% -------------------------------------------------------------------------
if any([MAP_options.return_log_PAH_MAP,MAP_options.return_log_OF_MAP,MAP_options.return_log_BFL_MAP])
    int_options.calc_log_PAH_MAP                                = true;
end
% -------------------------------------------------------------------------
% Return partial(log(P(A|Z,H)))/partial(inv_C_ee) at the MAP point
% -------------------------------------------------------------------------
if MAP_options.return_p_log_PAZH_p_inv_C_ee_MAP
    int_options.calc_p_log_PAZH_p_inv_C_ee_MAP                  = true;
end
% -------------------------------------------------------------------------
% Return partial(det(inv_C_aa))/partial(inv_C_ee) at the MAP point
% -------------------------------------------------------------------------
if MAP_options.return_p_det_inv_C_aa_p_inv_C_ee_MAP
    int_options.calc_p_det_inv_C_aa_p_inv_C_ee_MAP              = true;
end
% -------------------------------------------------------------------------
% Return partial(det(inv_C_aa))/partial(a) at the MAP point
% -------------------------------------------------------------------------
if MAP_options.return_p_det_inv_C_aa_p_a_MAP
    int_options.calc_p_det_inv_C_aa_p_a_MAP                     = true;
end
% -------------------------------------------------------------------------
% Return d(a_MAP)/d(inv_C_ee) at the MAP point
% -------------------------------------------------------------------------
if MAP_options.return_da_MAP_dinv_C_ee_MAP
    int_options.calc_inv_C_aa_MAP                               = true;
    int_options.calc_da_MAP_dinv_C_ee_MAP                       = true;
end
% -------------------------------------------------------------------------
% Return log(ML) at the MAP point (Marginal Likelihood)
% -------------------------------------------------------------------------
if MAP_options.return_log_ML_MAP
    int_options.calc_inv_C_aa_MAP                               = true;
    int_options.calc_logPAZH_MAP                                = true;
    int_options.calc_log_ML_MAP                                 = true;
end
% -------------------------------------------------------------------------
% Return d(log(ML))/d(inv_C_ee) at the MAP point
% -------------------------------------------------------------------------
if MAP_options.return_d_log_ML_MAP_d_inv_C_ee
    int_options.calc_p_log_PAZH_p_inv_C_ee_MAP                  = true;
    int_options.calc_p_det_inv_C_aa_p_inv_C_ee_MAP              = true;
    int_options.calc_p_det_inv_C_aa_p_a_MAP                     = true;
    int_options.calc_da_MAP_dinv_C_ee_MAP                       = true;
    int_options.calc_d_log_ML_MAP_d_inv_C_ee                    = true;
end
% -------------------------------------------------------------------------
% Say whether inv_C_ee (a.k.a. d2JdAdA) should be calculated exactly
% -------------------------------------------------------------------------
JdJd2J_options.d2JdAdA_exact = MAP_options.force_exact_inv_C_aa_MAP;
% -------------------------------------------------------------------------
% Normalize the Gaussians by default. 
% Note: The only time the Gaussians do not need to be normalized is when
% converging to a_MAP without varying the measurement noise. I could
% switch normalize_Gaussians = false for this state but it is a risk.
% Therefore, at least for now, I always normalize the Gaussians. 
% -------------------------------------------------------------------------
JdJd2J_options.normalize_Gaussians                              = true;

%% ========================================================================
% Either re-converge to a_MAP or re-calculate J, dJ, d2J at a_init
% =========================================================================
if int_options.reconverge
    % ---------------------------------------------------------------------
    % Set options for re-convergence to a_MAP, at which P(A|Z,H) is maximized
    % ---------------------------------------------------------------------
    % Set options that are required to maximize the posterior likelihood, PAZH
    JdJd2J_options.include_prior                                = true;
    JdJd2J_options.include_z_in_data_likelihood                 = true;
    % Set options that are not required to maximize the posterior likelihood
    JdJd2J_options.d_det_d2JdAdA_d_inv_C_ee                     = false;
    JdJd2J_options.d_det_d2JdAdA_dA                             = false;
    % ---------------------------------------------------------------------
    % Set options for re-convergence to a_MAP if a is being varied
    % ---------------------------------------------------------------------
    if MAP_options.reconverge_to_a_MAP_varying_a
        % Tell JdJd2J that we need the derivatives of J w.r.t. A
        JdJd2J_options.by_dA                                    = true;
        % Set the starting point for a (this is set to a_prior but does not have to be)
        AZQ = a_init;
        % Find the length of a so that a can be extracted from AZQ later
        N_a = length(a_init);
    else
        % Tell JdJd2J that we do not need the derivatives of J w.r.t. A
        JdJd2J_options.by_dA                                    = false;
        % Initialize AZQ
        AZQ = [];
        % Set N_a to 0
        N_a = 0;
    end
    % ---------------------------------------------------------------------
    % Set options for re-convergence to a_MAP if inv_C_ee is being varied
    % ---------------------------------------------------------------------
    if MAP_options.reconverge_to_a_MAP_varying_inv_C_ee
        % Tell JdJd2J that we need the derivatives of J w.r.t. inv_C_ee
        JdJd2J_options.by_dinv_C_ee                             = true;
        % Add the starting point for inv_C_ee to AZQ
        AZQ = [AZQ ; inv_C_ee_init];
        % Find the length of inv_C_ee so that it can be extracted from AZQ later
        N_z = length(inv_C_ee_init);
    else
        % Tell JdJd2J that we do not need the derivatives of J w.r.t. inv_C_ee
        JdJd2J_options.by_dinv_C_ee                             = false;
        % Set N_Z = 0
        N_z = 0;
    end
    %
    switch MAP_options.optim_method
        case 'Newton'
            if MAP_options.verbose; disp('Using Newton method'); end
            % Set dummy delAZQ
            delAZQ = 2*MAP_options.tol*ones(1,N_a+N_z);
            % Tell JdJd2J what needs to be calculated
            if int_options.calc_logPAZH_MAP
                JdJd2J_options.calculate_J                      = true;
            else
                JdJd2J_options.calculate_J                      = false;
            end
            JdJd2J_options.calculate_dJ_                        = true;
            JdJd2J_options.calculate_d2J_                       = true;
            while norm(abs(delAZQ)) > MAP_options.tol
                % Display to screen
                if MAP_options.verbose
                    fprintf([repmat('%-+10.4f  ',1,N_a+N_z),'\n'],AZQ.')
                end
                % Unpack AZQ
                [a_Newton,inv_C_ee_Newton] = fun_unpack_AZQ(MAP_options,a_init,inv_C_ee_init,AZQ,N_a,N_z);
                % Calculate dJ and d2J
                [J,dJ,d2J] = fun_JdJd2J(JdJd2J_options,param,a_ref,a_Newton,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee_Newton,z);
                % Update delaC = - dJdaC * inv(d2JdaCdaC) = - inv(d2JdaCdaC) * dJdaC.'
                delAZQ = - d2J \ dJ.';
                % Update aC
                AZQ = AZQ + delAZQ;
            end
        case 'BFGS_dJd2J'
            % Minimize the negative log likelihood using a BFGS gradient-based optimization algorithm with Hessian information
            if MAP_options.verbose; disp('Using BFGS with Jacobian and Hessian information'); end
            % Tell fun_JdJd2J what needs to be calculated
            JdJd2J_options.calculate_J                          = true; % required for fminunc
            JdJd2J_options.calculate_dJ_                        = true; % required for fminunc
            JdJd2J_options.calculate_d2J_                       = true; % required for fminunc
            % Set options for fminunc
            options = optimoptions('fminunc','Algorithm','trust-region','SpecifyObjectiveGradient',true,'HessianFcn','objective','OptimalityTolerance',MAP_options.tol); % default
            % Switch off output
            if ~MAP_options.verbose; options = optimoptions('fminunc','Display','off'); end
            % Find the posterior of AZQ and return the penultimate d2J
            [AZQ,J,~,~,~,d2J] = fminunc(@fun_JdJd2J_internal,AZQ,options);
        case 'BFGS_dJ'
            % Minimize the negative log likelihood using a BFGS gradient-based optimization algorithm without Hessian information
            if MAP_options.verbose; disp('Using BFGS with Jacobian (but not Hessian) information'); end
            % Tell fun_JdJd2J what needs to be calculated
            JdJd2J_options.calculate_J                          = true; % required for fminunc
            JdJd2J_options.calculate_dJ_                        = true; % required for fminunc
            JdJd2J_options.calculate_d2J_                       = false;
            % Set options for fminunc
            options = optimoptions('fminunc','Algorithm','trust-region','SpecifyObjectiveGradient',true,'OptimalityTolerance',MAP_options.tol); % default
            % Switch off output
            if ~MAP_options.verbose; options = optimoptions('fminunc','Display','off'); end
            % Find the posterior of AZQ and return the approximate value of d2J found with finite difference during the BFGS algorithm
            [AZQ,J,~,~,~,hessian] = fminunc(@fun_JdJ_internal,AZQ,options); d2J = full(hessian);
        otherwise
            disp('ERROR: assim_options.optim_method not recognized')
            return
    end
    % Unpack AZQ
    [a_MAP,inv_C_ee_MAP] = fun_unpack_AZQ(MAP_options,a_init,inv_C_ee_init,AZQ,N_a,N_z);
else
    % ---------------------------------------------------------------------
    % Calculate J, dJ, and d2J (as required) at a_init
    % ---------------------------------------------------------------------
    a_MAP =        a_init ; inv_C_ee_MAP =        inv_C_ee_init;
    N_a   = length(a_init);   N_z        = length(inv_C_ee_init);
    % Set the options if log(P(A|Z,H)) is required
    if int_options.calc_logPAZH_MAP
        JdJd2J_options.calculate_J                              = true;
    else
        JdJd2J_options.calculate_J                              = false;
    end
    % Set the options if inv_C_aa is required
    if int_options.calc_inv_C_aa_MAP
        JdJd2J_options.calculate_d2J_                           = true;
        JdJd2J_options.by_dA                                    = true;
        JdJd2J_options.by_dinv_C_ee                             = false;
    else
        JdJd2J_options.calculate_d2J_                           = false;
    end
    % Set the options if the code should check that a_init is a MAP point
    if int_options.check_is_MAP_point
        JdJd2J_options.calculate_dJ_                            = true;
        JdJd2J_options.by_dA                                    = true;
        JdJd2J_options.by_dinv_C_ee                             = false;
    else
        JdJd2J_options.calculate_dJ_                            = false;
    end
    if any([JdJd2J_options.calculate_J,JdJd2J_options.calculate_dJ_,JdJd2J_options.calculate_d2J_])
        % Set options that are required to maximize the posterior likelihood, PAZH
        JdJd2J_options.include_prior                            = true;
        JdJd2J_options.include_z_in_data_likelihood             = true;
        % Set options that are not required to maximize the posterior likelihood
        JdJd2J_options.d_det_d2JdAdA_d_inv_C_ee                 = false;
        JdJd2J_options.d_det_d2JdAdA_dA                         = false;
        [J,dJ,d2J] = fun_JdJd2J(JdJd2J_options,param,a_ref,a_init,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee_init,z);
        % Check that dJ/dA is small
        if norm(dJ) > 1e-6
            disp('WARNING: a_init is NOT a MAP point')
        end
    end
end

% -------------------------------------------------------------------------
% Embedded (local) function for BFGS using Jacobian and Hessian
    function [J,dJ,d2J] = fun_JdJd2J_internal(AZQ_local)
        % Display to screen
        if MAP_options.verbose; fprintf([repmat('%-+10.4f  ',1,N_a+N_z),'\n'],AZQ_local.'); end
        % Unpack AZQ_local
        [a_local,inv_C_ee_local] = fun_unpack_AZQ(MAP_options,a_init,inv_C_ee_init,AZQ_local,N_a,N_z);
        % Calculate J and dJ/d(AZQ) and d2(J)/d(AZQ)^2
        [J,dJ,d2J] = fun_JdJd2J(JdJd2J_options,param,a_ref,a_local,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee_local,z);
    end

% Embedded (local) function for BFGS using Jacobian (but not Hessian)
    function [J,dJ] = fun_JdJ_internal(AZQ_local)
        % Display to screen
        if MAP_options.verbose; fprintf([repmat('%-+10.4f  ',1,N_a+N_z),'\n'],AZQ_local.'); end
        % Unpack AZQ_local
        [a_local,inv_C_ee_local] = fun_unpack_AZQ(MAP_options,a_init,inv_C_ee_init,AZQ_local,N_a,N_z);
        % Calculate J and dJ/d(AZQ)
        [J,dJ]      = fun_JdJd2J(JdJd2J_options,param,a_ref,a_local,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee_local,z);
    end
% -------------------------------------------------------------------------

% -------------------------------------------------------------------------
% Extract inv_C_aa_MAP
% -------------------------------------------------------------------------
if int_options.calc_inv_C_aa_MAP
    inv_C_aa_MAP = d2J(1:N_a,1:N_a);
else
    inv_C_aa_MAP = [];
end

% -------------------------------------------------------------------------
% Extract log{ P(A_MAP|ZH) }
% -------------------------------------------------------------------------
if int_options.calc_logPAZH_MAP
    log_PAZH_MAP = - J;
else
    log_PAZH_MAP = [];
end

%% ======================================================================== 
% Calculate d(a_MAP)/d(inv_C_ee)
% =========================================================================
%
% When calculating d(log_ML)/d(inv_C_ee) we will need to calculate
% d(a_MAP)/d(inv_C_ee). This tells us how a_MAP changes when inv_C_ee
% changes. I explain the principle behind this in 2D, for simplicity,
% before introducing the general result. 
%
% -------------------------------------------------------------------------
% Simple explanation
% -------------------------------------------------------------------------
% Consider a function J(a,q) where a and q are scalars (q = inv_C_ee)
%
% The point a_MAP is defined such that p(J)/p(a)|q = 0 
% (i.e. partial(J)/partial(a), keeping q constant = 0)
%
% We want to calculate d(a_MAP)/d(q) 
%
% Imagine moving from point A to point B to point C
%
%  q                    Points A and C have p(J)/p(a)|q = 0 
%  |                    Point B is an interim point where p(J)/p(a)|q ~= 0 
%  |                    
%  |                    
%  |     B---C          
%  |     |                  
%  |     A              
%  |                     
% ---------------> a    
%  
%                         p(J)|
% From A to B: delta(J) = ----|    * delta(q)
%                         p(q)|a,A
%
%                         p(J)|
% From B to C: delta(J) = ----|    * delta(a)
%                         p(a)|q,B
%
%                         p(J)|                 p(J)| 
% Therefore J(C) = J(A) + ----|    * delta(q) + ----|    * delta(a) [1] 
%                         p(q)|a,A              p(a)|q,B
%
% But A and C are both MAP points, at which p(J)/p(a)|q. Therefore we take
% partial()/partial(a)|q of both sides of [1] to get:
%
%           p2(J) |                   p2(J) | 
% 0 = 0 + --------|    * delta(q) + --------|    * delta(a)
%         p(a)p(q)|a,A              p(a)p(a)|q,B
%
% Re-arranging and setting delta(a) and delta(q) to small values we get:
%
%  da     p2(J)/p(a)p(q)
%  -- = - --------------
%  dq     p2(J)/p(a)p(a)
%
% -------------------------------------------------------------------------
% General explanation, as implemented below
% -------------------------------------------------------------------------
% Now consider the problem in J(a,q) space, where a and q are col vectors
%
% Define:
%    grad_a = e_i * p(.)/p(a_i) where e_i is the i^th direction in a-space
%    grad_q = e_k * p(.)/p(q_k) where e_k is the k^th direction in q-space
%
% For any q, the MAP point is defined as 'a' for which grad_a(J) = 0.
%
% If q -> q + deltaq then J -> J + deltaJ where:
% 
%    deltaJ = grad_a(J) dot deltaa  +  grad_q(J) dot deltaq
%
%                  p(J)                     p(J)
%           = e_i ------ dot e_j deltaa_j + e_k ------ dot e_l deltaq_l
%                 p(a_i)                   p(q_k)
%
% Note that deltaa and deltaq are not functions of a or q, so:
%
%              p(J)              p(J)
%    deltaJ = ------ deltaa_j + ------ deltaq_l
%             p(a_j)            p(q_l)
%
% We are given deltaq_l and want to find the new MAP point, which occurs at
% deltaa_j such that grad_a(J + deltaJ) = 0, i.e. such that grad_a(deltaJ) = 0:
%
%                           p    {  p(J)              p(J)           }
%    grad_a(deltaJ) = e_i ------ { ------ deltaa_j + ------ deltaq_l } = 0
%                         p(a_i) { p(a_j)            p(q_l)          }
%
%        {     p2(J)                    p2(J)              }
% => e_i { ------------ deltaa_j  +  ------------ deltaq_l } = 0 for all e_i  
%        { p(a_i)p(a_j)              p(a_i)p(q_l)          }
%
% This represents I equations (one for each e_i) for I unknowns (deltaa_j),
% so can (usually) be solved. It requires a matrix formulation:
%   
%  [      ]     [              ]^-1     [              ]
%  [      ]     [    p2(J)     ]        [    p2(J)     ] [      ]
%  [deltaa] = - [ ------------ ]    *   [ ------------ ] [deltaq]
%  [      ]     [ p(a_i)p(a_j) ]        [ p(a_i)p(q_l) ] [      ]
%  [      ]     [              ]        [              ]
%
% -------------------------------------------------------------------------
% TECHNICAL NOTE
%   At the moment there is no way to tell fun_JdJd2J that 
%    we want p2(J)/p(a)p(a) (which is already in memory)
%    and     p2(J)/p(a)p(inv_C_ee) 
%    but not p2(J)/p(inv_C_ee)p(inv_C_ee). 
%   Therefore we calculate all of the above and strip out the first two
%   afterwards. To make the calculation cheaper, switch off d2JdAdA_exact
% -------------------------------------------------------------------------
% Calculate d(a_MAP)/d(inv_C_ee)
% -------------------------------------------------------------------------
if int_options.calc_da_MAP_dinv_C_ee_MAP
    % p2(J)/p(a)p(a) should already be in memory as inv_C_aa_MAP
    p2Jpapa = inv_C_aa_MAP;
    % Switch off exact d2JdAdA calculation because it is only required for p2Jpapa, which has already been calcualted 
    JdJd2J_options.d2JdAdA_exact                                = false;
    % Set other options to calculate p2(J)/p(a)p(inv_C_ee)
    JdJd2J_options.calculate_J                                  = false;
    JdJd2J_options.calculate_dJ_                                = false;
    JdJd2J_options.calculate_d2J_                               = true;
    JdJd2J_options.by_dinv_C_ee                                 = true;
    JdJd2J_options.by_dA                                        = true;
    % Calculate 'approximate p2Jpapa', which we throw away, and p2Jpapq
    [~,~,d2J] = fun_JdJd2J(JdJd2J_options,param,a_ref,a_MAP,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee_MAP,z);
    % Re-count the number of noise parameters (this will not necessarily have been done before) 
    N_z = length(inv_C_ee_MAP);
    % Extract p2(J)/p(a_i)p(q_l)
    p2Jpapq = d2J(1:N_a,N_a+1:N_a+N_z);
    % Calculate d(a)/d(inv_C_ee) at the MaxL point
    da_MAP_dinv_C_ee_MAP = - p2Jpapa \ p2Jpapq; % - inv(p2Jpapa) * p2Jpapq;
    % Switch exact d2JdAdA to what it was before
    JdJd2J_options.d2JdAdA_exact = MAP_options.force_exact_inv_C_aa_MAP;
else 
    da_MAP_dinv_C_ee_MAP = [];
end

%% ========================================================================
% CALCULATE THE MARGINAL LIKELIHOOD AND THE BUILDING BLOCKS OF ITS GRADIENTS 
% ========================================================================
%
% The Marginal Likelihood, P(Z|H), can be calculated with Laplace's method:
%
%   Posterior likelihood: P(A_MAP|Z,H) including normalizations
%                            |
%           <----------------------->
%  P(Z|H) ~ P(Z|A_MAP,H) * P(A_MAP|H) * 1/sqrt(det(inv_C_aa_MAP/2pi))   [1]
%     |            |       <---------------------------------------->
%     |            |                           |
%  Evidence  Best fit likelihood          Occam factor
%
% in the notation in this file, this is:
%   log_ML_MAP = log_PAZH_MAP - 0.5*log(det(inv_C_aa_MAP)) + 0.5*N_a*log(2*pi) 
% 
% log_ML is a differentiable function of (a_p,inv_C_aa,z_data,inv_C_ee). 
% It is obvious that the optimum would be found by setting a_p = a_MAP and
% inv_C_aa --> infty. However, it makes no sense to change a_p or 
% inv_C_aa AFTER we have assimilated the data because they are priors. 
% Further, z_data is fixed, so we should not change this. 
% This leaves inv_C_ee as the only variable that should be changed:
%
% d(log_ML_MAP)   d(log_PAZH_MAP)              1              d(det(inv_C_aa_MAP))         .
% ------------- = --------------- - 0.5 * ----------------  * --------------------         .
%  d(inv_C_ee)      d(inv_C_ee)           det(inv_C_aa_MAP)       d(inv_C_ee)             .
%
% where:
%
% d(log_PAZH_MAP)   p(log_PAZH_MAP)   p(log_PAZH_MAP)     d(a_MAP)
% --------------- = --------------- + --------------- * -----------
%   d(inv_C_ee)       p(inv_C_ee)          p(a)         d(inv_C_ee)
%                                           |
%                                 This term is zero at the MAP
% and
%
% d(det(d2Jdada_MAP))   p(det(d2Jdada_MAP))   p(det(d2Jdada_MAP))    d(a_MAP)
% ------------------- = ------------------- + ------------------- * -----------
%     d(inv_C_ee)           p(inv_C_ee)               p(a)          d(inv_C_ee)
%
% -------------------------------------------------------------------------
% Calculate log(Marginal Likelihood) = log_ML = log(PZH) at the MAP point
% -------------------------------------------------------------------------
if int_options.calc_log_ML_MAP
    log_ML_MAP = log_PAZH_MAP - 0.5 * log(det(inv_C_aa_MAP)) + 0.5 * N_a * log(2*pi);
else
    log_ML_MAP = [];
end
% -------------------------------------------------------------------------
% Calculate p(log_PAZH_MAP)/p(inv_C_ee)
% -------------------------------------------------------------------------
if int_options.calc_p_log_PAZH_p_inv_C_ee_MAP
    % If inv_C_ee has been varied to find the MAP point then partial(log_PAZH)/partial(inv_C_ee) should be zero to the optimization tolerance
    % Calculate partial(log_PAZH)/partial(inv_C_ee) at the MAP point
    JdJd2J_options.calculate_J                  = false;
    JdJd2J_options.calculate_dJ_                = true;
    JdJd2J_options.calculate_d2J_               = false;
    JdJd2J_options.normalize_Gaussians          = true;
    JdJd2J_options.by_dA                        = false;
    JdJd2J_options.by_dinv_C_ee                 = true;  % <- switch on gradient w.r.t. inv_C_ee
    JdJd2J_options.d_det_d2JdAdA_d_inv_C_ee     = false;
    JdJd2J_options.d_det_d2JdAdA_dA             = false;
    JdJd2J_options.include_prior                = true;  % <- include prior
    JdJd2J_options.include_z_in_data_likelihood = true;  % <- include likelihood
    [~,dJ,~] = fun_JdJd2J(JdJd2J_options,param,a_ref,a_MAP,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee_MAP,z);
    p_log_PAZH_p_inv_C_ee_MAP = - dJ;
else
    p_log_PAZH_p_inv_C_ee_MAP = [];
end
% -------------------------------------------------------------------------
% Calculate p(det(d2Jdada_MAP))/p(inv_C_ee) and p(det(d2Jdada_MAP))/p(A)
% -------------------------------------------------------------------------
if or(int_options.calc_p_det_inv_C_aa_p_inv_C_ee_MAP,int_options.calc_p_det_inv_C_aa_p_a_MAP)
    JdJd2J_options.include_prior                = true;  % <- include prior
    JdJd2J_options.include_z_in_data_likelihood = true;  % <- include likelihood
    JdJd2J_options.calculate_J                  = false;
    JdJd2J_options.calculate_dJ_                = false;
    JdJd2J_options.calculate_d2J_               = true;
    JdJd2J_options.normalize_Gaussians          = true;
    JdJd2J_options.by_dA                        = true;  % <- switch on gradient w.r.t. a to get d2J/dada
    JdJd2J_options.by_dinv_C_ee                 = false; % <- switch off gradient w.r.t. inv_C_ee to avoid d2J/d(inv_C)/d(inv_C)
    JdJd2J_options.d2JdAdA_exact                = true;
    JdJd2J_options.d_det_d2JdAdA_d_inv_C_ee     = true;  % <- switch on p(det(d2Jdada_MAP))/p(inv_C_ee)
    JdJd2J_options.d_det_d2JdAdA_dA             = true;  % <- switch on p(det(d2Jdada_MAP))/p(a)
    [~,~,~,p_det_inv_C_aa_p_inv_C_ee_MAP,p_det_inv_C_aa_p_a_MAP] = fun_JdJd2J(JdJd2J_options,param,a_ref,a_MAP,a_prior,inv_C_aa_prior,C_ee_ref,inv_C_ee_MAP,z);
else
    p_det_inv_C_aa_p_inv_C_ee_MAP = []; p_det_inv_C_aa_p_a_MAP = [];
end
% -------------------------------------------------------------------------
% Assemble these into d(log_ML)/d(inv_C_ee) at the MAP point
% -------------------------------------------------------------------------
if int_options.calc_d_log_ML_MAP_d_inv_C_ee
    % Calculate d_log_PAZH_d_inv_C_ee_MAP (internal variable, not for export)
    d_log_PAZH_d_inv_C_ee_MAP     = p_log_PAZH_p_inv_C_ee_MAP;
    % Calculate d_det_inv_C_aa_d_inv_C_ee_MAP (internal variable, not for export)
    d_det_inv_C_aa_d_inv_C_ee_MAP = p_det_inv_C_aa_p_inv_C_ee_MAP ...
                                  + p_det_inv_C_aa_p_a_MAP.' * da_MAP_dinv_C_ee_MAP;
    % Calculate d_log_ML_MAP_d_inv_C_ee    
    d_log_ML_MAP_d_inv_C_ee       = d_log_PAZH_d_inv_C_ee_MAP ...
                                  - 0.5 * d_det_inv_C_aa_d_inv_C_ee_MAP / det(inv_C_aa_MAP);
else
    d_log_ML_MAP_d_inv_C_ee = [];
end

% -------------------------------------------------------------------------
% Calculate the prior P(A|Z) (for safety, do this last because funJdJd2J.include_z_in_data_likelihood is set to false)
% -------------------------------------------------------------------------
if int_options.calc_log_PAH_MAP
    JdJd2J_options.calculate_J                      = true;
    JdJd2J_options.calculate_dJ_                    = false;
    JdJd2J_options.calculate_d2J_                   = false;
    JdJd2J_options.include_z_in_data_likelihood     = false;
    JdJd2J_options.d_det_d2JdAdA_d_inv_C_ee         = false;
    JdJd2J_options.d_det_d2JdAdA_dA                 = false;
    J = fun_JdJd2J(JdJd2J_options,[],a_ref,a_MAP,a_prior,inv_C_aa_prior,[],[],[]);
    log_PAH_MAP = - J;
else
    log_PAH_MAP = [];
end

% -------------------------------------------------------------------------
% Calculate the Best Fit Likelihood and the Occam Factor (cheap)
% -------------------------------------------------------------------------
log_BFL = log_PAZH_MAP - log_PAH_MAP;
log_OF  = log_ML_MAP - log_BFL;

end

%% Utility functions

function [a,inv_C_ee] = fun_unpack_AZQ(MAP_options,a_init,inv_C_ee_init,AZQ,N_a,N_Z)
% Unpack the vector being optimized, AZQ, into A, Z, and Q

% If the parameters, a, are being varied then extract a from AZQ. Otherwise use a_init
if MAP_options.reconverge_to_a_MAP_varying_a
    a = AZQ(1:N_a);
else
    a = a_init;
end

% If the measurement noise, inv_C_ee, is being varied then extract from AZQ. Otherwise use inv_C_ee_init 
if MAP_options.reconverge_to_a_MAP_varying_inv_C_ee
    inv_C_ee = AZQ(N_a+1:N_a+N_Z);
else
    inv_C_ee = inv_C_ee_init;
end

end

