function [s_post,V_s_post] = fun_covar_s_P_a(param,a_ref,a_post,inv_C_aa_post)
%
% For a single configuration, calculate the posterior STATE (s & P)
% expected values and variances from the posterior PARAMETER (a) expected
% values and covariance matrix.  
% 
% INPUTS
%   param           the parameters of the configuration being examined
%   a_ref           the reference values of the parameters
%   a_post          the posterior expected value of the parameters
%   C_aa_post       the posterior covariance matrix of the parameters 
%
% OUTPUTS
%   sr_post         posterior expected value of the growth rate
%   si_post         posterior expected value of the frequency
%   Pr_post         posterior expected value of the real pressure
%   Pi_post         posterior expected value of the imag pressure
%   V_sr_post       posterior variance of sr
%   V_si_post       posterior variance of si
%   V_Pr_post       posterior variance of Pr
%   V_Pi_post       posterior variance of Pi
%
% Having calculated the posterior covariance of the parameters, C_aa_post, we have:
%   p(a|z) ~ exp{ - 0.5 * (a - a0)' * inv(C_aa_post) * (a - a0) }    [when a is arranged as a column vector]
% where a0 is the most likely set of parameters and Caa_a is the posterior
% covariance of the parameters around this solution.
%
% We want to find the corresponding confidence intervals in the model
% predictions. A clear explanation of how to do this is in  4.3 of Hans
% Yu's thesis, which I have paraphrased here. 
%
% We want to find a for which |s(a) - s(a0)|^2 is maximized, subject to:
% (a - a0)' * inv(C_aa_post) * (a - a0) = r^2, where r denotes the
% confidence level in units of standard deviations. 
%
% A first order approximation of the above gives:
%  |s(a) - s(a0)|^2 ~ (a - a0)' * [dsda * dsda'] * (a - a0)
%                                 <-outer prod->
%
% The method of Lagrange multipliers gives the necessary condition:
%  [dsda * dsda'] * (a - a0) = lambda * inv(C_aa_post) * (a - a0)
% 
% This is a generalized eigenvalue problem for lambda:
%  [lambda , (a - a0)] = eig( [dsda * dsda'] , inv(C_aa_post) )
%
% The standard deviation of s is given by the maximum eigenvalue, lambda

% Enter the expected values of the parameters into the vaparam vector
param.varparams = a_ref.*a_post;
% Calculate s and dsda at these param values
[s_post,dsda] = fun_solve_s(param);
% Rescale inv_C_aa_post by a_ref
inv_C_aa_post = inv_C_aa_post ./ (a_ref * a_ref.');
% Calculate the variances
V_s_post = max(eig(dsda * dsda.',inv_C_aa_post));

end

