function [s,dsda,d2sdada] = fun_solve_s(param)
% ----------------------------------------------------------------------
% DESCRIPTION
% 
% This function contains the model, H(a), which has parameters, a, and
% produces output, s, and the first and second derivatives of s w.r.t. a.
%
% Here, the model is a standard polynomial. Usually, this function
% would contain a physics-based model.
%
% ----------------------------------------------------------------------
% NOMENCLATURE
%  a        parameter vector [ a(1) , ... , a(N) ]; a = param.varparams
%  N        number of parameters; i.e. N = length(a)
%  N-1      order of the polynomial
%  s        model output
%  dsda     derivative of s w.r.t. each element of a
%  d2sdada  second derivative of s w.r.t. pairs of elements of a
%
% The polynomial has order N-1 so:
%  s    =   a(0)*x^(N-1) + a(1)*x^(N-2) + ... + a(N) * x^0
%
% By inspection:
%  dsda = [      x^(N-1) ;      x^(N-2) ; ... ;        x^0 ];
%
% By inspection:
%  d2sdada = N x N zero matrix; zeros(N)
%
% ----------------------------------------------------------------------

% Evaluate s
s = polyval(param.varparams,param.x);
% Find the length of a
N = length(param.varparams);
% Evaluate dsda
method = 'vectorized';
switch method
    case 'serial'
        % Evaluate dsda with a loop (left in for clarity) 
        dsda = NaN(N,1);
        for n = 1:N
            dsda(n) = param.x^(N-n);
        end
    case 'vectorized'
        % Evaluate dsda in one line (quicker)
        dsda = param.x.^(N-(1:N)');
end
% Evaluate d2sdada
d2sdada = zeros(N);

end
